import logging
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from typing import Union, List, ClassVar, DefaultDict, Set, Literal, Annotated

import pytest

from dataclass_wizard.properties import property_wizard
from .._typing import PY310_OR_ABOVE

log = logging.getLogger(__name__)


def test_property_wizard_does_not_affect_normal_properties():
    """
    The `property_wizard` should not otherwise affect normal properties (i.e. ones
    that don't have their property names (or underscored names) annotated as a
    dataclass field.

    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        def __post_init__(self):
            self.wheels = 4
            self._my_prop = 0

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

        @property
        def _my_prop(self) -> int:
            return self.my_prop

        @_my_prop.setter
        def _my_prop(self, my_prop: Union[int, str]):
            self.my_prop = int(my_prop) + 5

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 4
    assert v._my_prop == 5

    # These should all result in a `TypeError`, as neither `wheels` nor
    # `_my_prop` are valid arguments to the constructor, as they are just
    # normal properties.

    with pytest.raises(TypeError):
        _ = Vehicle(wheels=3)

    with pytest.raises(TypeError):
        _ = Vehicle('6')

    with pytest.raises(TypeError):
        _ = Vehicle(_my_prop=2)

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'

    v._my_prop = '5'
    assert v._my_prop == 10, 'Expected assignment to use the setter method'


def test_property_wizard_does_not_affect_read_only_properties():
    """
    The `property_wizard` should not otherwise affect properties which are
    read-only (i.e. ones which don't define a `setter` method)

    """
    @dataclass
    class Vehicle(metaclass=property_wizard):
        list_of_wheels: list = field(default_factory=list)

        @property
        def wheels(self) -> int:
            return len(self.list_of_wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    # AttributeError: can't set attribute
    with pytest.raises(AttributeError):
        v.wheels = 3

    v = Vehicle(list_of_wheels=[1, 2, 1])
    assert v.wheels == 3

    v.list_of_wheels = [0]
    assert v.wheels == 1


def test_property_wizard_does_not_error_when_forward_refs_are_declared():
    """
    Using `property_wizard` when the dataclass has a forward reference
    defined in a type annotation.

    """
    @dataclass
    class Car:
        tires: int

    @dataclass
    class Truck:
        color: str

    globals().update(locals())

    @dataclass
    class Vehicle(metaclass=property_wizard):

        fire_truck: 'Truck'
        cars: List['Car'] = field(default_factory=list)

        _wheels: Union[int, str] = 4

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    truck = Truck('red')

    v = Vehicle(fire_truck=truck)
    log.debug(v)
    assert v.wheels == 4

    v = Vehicle(fire_truck=truck, wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle(truck, [Car(4)], '6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_public_property_and_underscored_field():
    """
    Using `property_wizard` when the dataclass has an public property and an
    underscored field name.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        _wheels: Union[int, str] = 4

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 4

    # Note that my IDE complains here, and suggests `_wheels` as a possible
    # keyword argument to the constructor method; however, that's wrong and
    # will error if you try it way.
    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_public_property_and_field():
    """
    Using `property_wizard` when the dataclass has both a property and field
    name *without* a leading underscore.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `wheels` here will be ignored, since `wheels` is simply
        # re-assigned on the following property definition.
        wheels: Union[int, str] = 4

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


@pytest.mark.skipif(not PY310_OR_ABOVE, reason='requires Python 3.10 or higher')
def test_property_wizard_with_public_property_and_field_with_or():
    """
    Using `property_wizard` when the dataclass has both a property and field
    name *without* a leading underscore, and using the OR ("|") operator in
    Python 3.10+, instead of the `typing.Union` usage.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `wheels` here will be ignored, since `wheels` is simply
        # re-assigned on the following property definition.
        wheels: int | str = 4

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_underscored_property_and_public_field():
    """
    Using `property_wizard` when the dataclass has an underscored property and
    a public field name.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: Union[int, str] = 4

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 4

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_underscored_property_and_field():
    """
    Using `property_wizard` when the dataclass has both a property and field
    name with a leading underscore.

    Note: this approach is generally *not* recommended, because the IDE won't
    know that the property or field name will be transformed to a public field
    name without the leading underscore, so it won't offer the desired type
    hints and auto-completion here.

    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `_wheels` here will be ignored, since `_wheels` is
        # simply re-assigned on the following property definition.
        _wheels: Union[int, str] = 4

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    # Note that my IDE complains here, and suggests `_wheels` as a possible
    # keyword argument to the constructor method; however, that's wrong and
    # will error if you try it way.
    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_public_property_and_annotated_field():
    """
    Using `property_wizard` when the dataclass has both a property and field
    name *without* a leading underscore, and the field is a
    :class:`typing.Annotated` type.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `wheels` here will be ignored, since `wheels` is simply
        # re-assigned on the following property definition.
        wheels: Annotated[Union[int, str], field(default=4)] = None

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 4

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_private_property_and_annotated_field_with_no_useful_extras():
    """
    Using `property_wizard` when the dataclass has both a property and field
    name with a leading underscore, and the field is a
    :class:`typing.Annotated` type without any extras that are a
    :class:`dataclasses.Field` type.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `wheels` here will be ignored, since `wheels` is simply
        # re-assigned on the following property definition.
        _wheels: Annotated[Union[int, str], 'Hello world!', 123] = None

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_multiple_inheritance():
    """
    When using multiple inheritance or when extending from more than one
    class, and if any of the super classes define properties that should also
    be `dataclass` fields, then the recommended approach is to define the
    `property_wizard` metaclass on each class that has such properties. Note
    that the last class in the below example (Car) doesn't need to use this
    metaclass, as it doesn't have any properties that meet this condition.

    """
    @dataclass
    class VehicleWithWheels(metaclass=property_wizard):
        _wheels: Union[int, str] = field(default=4)

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    @dataclass
    class Vehicle(VehicleWithWheels, metaclass=property_wizard):
        _windows: Union[int, str] = field(default=6)

        @property
        def windows(self) -> int:
            return self._windows

        @windows.setter
        def windows(self, windows: Union[int, str]):
            self._windows = int(windows)

    @dataclass
    class Car(Vehicle):
        my_list: List[str] = field(default_factory=list)

    v = Car()
    log.debug(v)
    assert v.wheels == 4
    assert v.windows == 6
    assert v.my_list == []

    # Note that my IDE complains here, and suggests `_wheels` as a possible
    # keyword argument to the constructor method; however, that's wrong and
    # will error if you try it way.
    v = Car(wheels=3, windows=5, my_list=['hello', 'world'])
    log.debug(v)
    assert v.wheels == 3
    assert v.windows == 5
    assert v.my_list == ['hello', 'world']

    v = Car('6', '7', ['testing'])
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'
    assert v.windows == 7, 'The constructor should use our setter method'
    assert v.my_list == ['testing']

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'
    v.windows = '321'
    assert v.windows == 321, 'Expected assignment to use the setter method'

# NOTE: the below test cases are added for coverage purposes


def test_property_wizard_with_public_property_and_underscored_field_without_default_value():
    """
    Using `property_wizard` when the dataclass has a public property, and an
    underscored field *without* a default value explicitly set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        _wheels: Union[int, str]

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_public_property_and_underscored_field_with_default_factory():
    """
    Using `property_wizard` when the dataclass has a public property, and an
    underscored field has only `default_factory` set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        _wheels: Union[int, str] = field(default_factory=str)

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    with pytest.raises(ValueError):
        # Setter raises ValueError, as `wheels` will be a string by default
        _ = Vehicle()

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_public_property_and_underscored_field_without_default_or_default_factory():
    """
    Using `property_wizard` when the dataclass has a public property, and an
    underscored field has neither `default` or `default_factory` set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        _wheels: Union[int, str] = field()

        @property
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_underscored_property_and_public_field_without_default_value():
    """
    Using `property_wizard` when the dataclass has an underscored property,
    and a public field *without* a default value explicitly set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: Union[int, str]

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_public_property_and_public_field_is_property():
    """
    Using `property_wizard` when the dataclass has an underscored property,
    and a public field is also defined as a property.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # The value of `wheels` here will be ignored, since `wheels` is simply
        # re-assigned on the following property definition.
        wheels = property
        # Defines the default value for `wheels`, since it won't work if we
        # define it above. The `init=False` is needed since otherwise IDEs
        # seem to suggest `_wheels` as a parameter to the constructor method,
        # which shouldn't be the case.
        #
        # Note: if are *ok* with the default value for the type (0 in this
        # case), then you can remove the below line and annotate the above
        # line instead as `wheels: Union[int, str] = property`
        _wheels: Union[int, str] = field(default=4, init=False)

        @wheels
        def wheels(self) -> int:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 4

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_underscored_property_and_public_field_with_default():
    """
    Using `property_wizard` when the dataclass has an underscored property,
    and the public field has `default` set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: Union[int, str] = field(default=2)

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 2

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_underscored_property_and_public_field_with_default_factory():
    """
    Using `property_wizard` when the dataclass has an underscored property,
    and the public field has only `default_factory` set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: Union[int, str] = field(default_factory=str)

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    with pytest.raises(ValueError):
        # Setter raises ValueError, as `wheels` will be a string by default
        _ = Vehicle()

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_underscored_property_and_public_field_without_default_or_default_factory():
    """
    Using `property_wizard` when the dataclass has an underscored property,
    and the public field has neither `default` or `default_factory` set.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: Union[int, str] = field()

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_where_annotated_type_contains_none():
    """
    Using `property_wizard` when the annotated type for the dataclass field
    associated with a property is here a :class:`Union` type that contains
    `None`. As such, the field is technically an `Optional` so the default
    value will be `None` if no value is specified via the constructor.

    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: Union[int, str, None]

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    # TypeError: int() argument is `None`
    with pytest.raises(TypeError):
        _ = Vehicle()

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('6')
    log.debug(v)
    assert v.wheels == 6, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_literal_type():
    """
    Using `property_wizard` when the dataclass field associated with a
    property is annotated with a :class:`Literal` type.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # Annotate `wheels` as a literal that should only be set to 1 or 0
        # (similar to how the binary numeral system works, for example)
        #
        # Note: we can assign a default value for `wheels` explicitly, so that
        # the IDE doesn't complain when we omit the argument to the
        # constructor method, but it's technically not required.
        wheels: Literal[1, '1', 0, '0']

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 1

    # The IDE should display a warning (`wheels` only accepts [0, 1]), however
    # it won't prevent the assignment here.
    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    # The IDE should display no warning here, as this is an acceptable value
    v = Vehicle('1')
    log.debug(v)
    assert v.wheels == 1, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_concrete_type():
    """
    Using `property_wizard` when the dataclass field associated with a
    property is annotated with a non-generic type, such as a `str` or `int`.
    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: int

        @property
        def _wheels(self) -> int:
            return self._wheels

        @_wheels.setter
        def _wheels(self, wheels: Union[int, str]):
            self._wheels = int(wheels)

    v = Vehicle()
    log.debug(v)
    assert v.wheels == 0

    v = Vehicle(wheels=3)
    log.debug(v)
    assert v.wheels == 3

    v = Vehicle('1')
    log.debug(v)
    assert v.wheels == 1, 'The constructor should use our setter method'

    v.wheels = '123'
    assert v.wheels == 123, 'Expected assignment to use the setter method'


def test_property_wizard_with_concrete_type_and_default_factory_raises_type_error():
    """
    Using `property_wizard` when the dataclass field associated with a
    property is annotated with a non-generic type, such as a `datetime`, which
    doesn't have a no-args constructor. Since `property_wizard` is not able to
    instantiate a new `datetime`, the default value should be ``None``.

    """
    @dataclass
    class Vehicle(metaclass=property_wizard):

        # Date when the vehicle was sold
        sold_dt: datetime

        @property
        def _sold_dt(self) -> int:
            return self._sold_dt

        @_sold_dt.setter
        def _sold_dt(self, sold_dt: datetime):
            """Save the datetime with the year set to `2010`"""
            self._sold_dt = sold_dt.replace(year=2010)

    # AttributeError: 'NoneType' object has no attribute 'replace'
    with pytest.raises(AttributeError):
        _ = Vehicle()

    dt = datetime(2020, 1, 1, 12, 0, 0)             # Jan. 1 2020 12:00 PM
    expected_dt = datetime(2010, 1, 1, 12, 0, 0)    # Jan. 1 2010 12:00 PM

    v = Vehicle(sold_dt=dt)
    log.debug(v)
    assert v.sold_dt != dt
    assert v.sold_dt == expected_dt, 'The constructor should use our setter ' \
                                     'method'

    dt = datetime.min
    expected_dt = datetime.min.replace(year=2010)

    v.sold_dt = dt
    assert v.sold_dt == expected_dt, 'Expected assignment to use the setter ' \
                                     'method'


def test_property_wizard_with_generic_type_which_is_not_supported():
    """
    Using `property_wizard` when the dataclass field associated with a
    property is annotated with a generic type other than one of the supported
    types (e.g. Literal and Union).

    """

    @dataclass
    class Vehicle(metaclass=property_wizard):
        # Date when the vehicle was sold
        sold_dt: ClassVar[datetime]

        @property
        def _sold_dt(self) -> int:
            return self._sold_dt

        @_sold_dt.setter
        def _sold_dt(self, sold_dt: datetime):
            """Save the datetime with the year set to `2010`"""
            self._sold_dt = sold_dt.replace(year=2010)

    v = Vehicle()
    log.debug(v)

    dt = datetime(2020, 1, 1, 12, 0, 0)  # Jan. 1 2020 12:00 PM
    expected_dt = datetime(2010, 1, 1, 12, 0, 0)  # Jan. 1 2010 12:00 PM

    # TypeError: __init__() got an unexpected keyword argument 'sold_dt'
    #   Note: This is expected because the field for the property is a
    #   `ClassVar`, and even `dataclasses` excludes this annotated type
    #   from the constructor.
    with pytest.raises(TypeError):
        _ = Vehicle(sold_dt=dt)

    # Our property should still work as expected, however
    v.sold_dt = dt
    assert v.sold_dt == expected_dt, 'Expected assignment to use the setter ' \
                                     'method'


def test_property_wizard_with_mutable_types_v1():
    """
    The `property_wizard` handles mutable collections (e.g. subclasses of list,
    dict, and set) as expected. The defaults for these mutable types should
    use a `default_factory` so we can observe the expected behavior.
    """

    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: List[Union[int, str]]
        # _wheels: List[Union[int, str]] = field(init=False)

        inverse_bool_set: Set[bool]
        # Not needed, but we can also define this as below if we want to
        # inverse_bool_set: Annotated[Set[bool], field(default_factory=set)]

        # We'll need the `field(default_factory=...)` syntax here, because
        # otherwise the default_factory will be `defaultdict()`, which is not what
        # we want.
        wheels_dict: Annotated[
            DefaultDict[str, List[str]],
            field(default_factory=lambda: defaultdict(list))
        ]

        @property
        def wheels(self) -> List[int]:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: List[Union[int, str]]):
            self._wheels = [int(w) for w in wheels]

        @property
        def inverse_bool_set(self) -> Set[bool]:
            return self._inverse_bool_set

        @inverse_bool_set.setter
        def inverse_bool_set(self, bool_set: Set[bool]):
            # Confirm that we're passed in the right type when no value is set via
            # the constructor (i.e. from the `property_wizard` metaclass)
            assert isinstance(bool_set, set)
            self._inverse_bool_set = {not b for b in bool_set}

        @property
        def wheels_dict(self) -> int:
            return self._wheels_dict

        @wheels_dict.setter
        def wheels_dict(self, wheels: Union[int, str]):
            self._wheels_dict = wheels

    v1 = Vehicle(wheels=['1', '2', '3'],
                 inverse_bool_set={True, False},
                 wheels_dict=defaultdict(list, key=['value']))
    v1.wheels_dict['key2'].append('another value')
    log.debug(v1)

    v2 = Vehicle()
    v2.wheels.append(4)
    v2.wheels_dict['a'].append('5')
    v2.inverse_bool_set.add(True)
    log.debug(v2)

    v3 = Vehicle()
    v3.wheels.append(1)
    v3.wheels_dict['b'].append('2')
    v3.inverse_bool_set.add(False)
    log.debug(v3)

    assert v1.wheels == [1, 2, 3]
    assert v1.inverse_bool_set == {False, True}
    assert v1.wheels_dict == {'key': ['value'], 'key2': ['another value']}

    assert v2.wheels == [4]
    assert v2.inverse_bool_set == {True}
    assert v2.wheels_dict == {'a': ['5']}

    assert v3.wheels == [1]
    assert v3.inverse_bool_set == {False}
    assert v3.wheels_dict == {'b': ['2']}


def test_property_wizard_with_mutable_types_v2():
    """
    The `property_wizard` handles mutable collections (e.g. subclasses of list,
    dict, and set) as expected. The defaults for these mutable types should
    use a `default_factory` so we can observe the expected behavior.

    In this version, we explicitly pass in the `field(default_factory=...)`
    syntax for all field properties, though it's technically not needed.
    """

    @dataclass
    class Vehicle(metaclass=property_wizard):
        wheels: Annotated[List[int], field(default_factory=list)]
        _wheels_list: list = field(default_factory=list)

        @property
        def wheels_list(self) -> list:
            return self._wheels_list

        @wheels_list.setter
        def wheels_list(self, wheels):
            self._wheels_list = wheels

        @property
        def wheels(self) -> list:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels):
            self._wheels = wheels

    v1 = Vehicle(wheels=[1, 2], wheels_list=[2, 1])
    v1.wheels.append(3)
    v1.wheels_list.insert(0, 3)
    log.debug(v1)

    v2 = Vehicle()
    log.debug(v2)

    v2.wheels.append(2)
    v2.wheels.append(1)
    v2.wheels_list.append(1)
    v2.wheels_list.append(2)

    v3 = Vehicle()
    log.debug(v3)

    v3.wheels.append(1)
    v3.wheels.append(1)
    v3.wheels_list.append(5)
    v3.wheels_list.append(5)

    assert v1.wheels == [1, 2, 3]
    assert v1.wheels_list == [3, 2, 1]
    assert v2.wheels == [2, 1]
    assert v2.wheels_list == [1, 2]
    assert v3.wheels == [1, 1]
    assert v3.wheels_list == [5, 5]


def test_property_wizard_with_mutable_types_with_parameterized_standard_collections():
    """
    Test case for mutable types with a Python 3.9 specific feature:
    parameterized standard collections. As such, this test case is only
    expected to pass for Python 3.9+.
    """

    @dataclass
    class Vehicle(metaclass=property_wizard):

        wheels: list[Union[int, str]]
        # _wheels: List[Union[int, str]] = field(init=False)

        inverse_bool_set: set[bool]
        # Not needed, but we can also define this as below if we want to
        # inverse_bool_set: Annotated[Set[bool], field(default_factory=set)]

        # We'll need the `field(default_factory=...)` syntax here, because
        # otherwise the default_factory will be `defaultdict()`, which is not what
        # we want.
        wheels_dict: Annotated[
            defaultdict[str, List[str]],
            field(default_factory=lambda: defaultdict(list))
        ]

        @property
        def wheels(self) -> List[int]:
            return self._wheels

        @wheels.setter
        def wheels(self, wheels: List[Union[int, str]]):
            self._wheels = [int(w) for w in wheels]

        @property
        def inverse_bool_set(self) -> Set[bool]:
            return self._inverse_bool_set

        @inverse_bool_set.setter
        def inverse_bool_set(self, bool_set: Set[bool]):
            # Confirm that we're passed in the right type when no value is set via
            # the constructor (i.e. from the `property_wizard` metaclass)
            assert isinstance(bool_set, set)
            self._inverse_bool_set = {not b for b in bool_set}

        @property
        def wheels_dict(self) -> int:
            return self._wheels_dict

        @wheels_dict.setter
        def wheels_dict(self, wheels: Union[int, str]):
            self._wheels_dict = wheels

    v1 = Vehicle(wheels=['1', '2', '3'],
                 inverse_bool_set={True, False},
                 wheels_dict=defaultdict(list, key=['value']))
    v1.wheels_dict['key2'].append('another value')
    log.debug(v1)

    v2 = Vehicle()
    v2.wheels.append(4)
    v2.wheels_dict['a'].append('5')
    v2.inverse_bool_set.add(True)
    log.debug(v2)

    v3 = Vehicle()
    v3.wheels.append(1)
    v3.wheels_dict['b'].append('2')
    v3.inverse_bool_set.add(False)
    log.debug(v3)

    assert v1.wheels == [1, 2, 3]
    assert v1.inverse_bool_set == {False, True}
    assert v1.wheels_dict == {'key': ['value'], 'key2': ['another value']}

    assert v2.wheels == [4]
    assert v2.inverse_bool_set == {True}
    assert v2.wheels_dict == {'a': ['5']}

    assert v3.wheels == [1]
    assert v3.inverse_bool_set == {False}
    assert v3.wheels_dict == {'b': ['2']}



"""
Helper Wizard Mixin classes.
"""
__all__ = ['JSONListWizard',
           'JSONFileWizard',
           'TOMLWizard',
           'YAMLWizard']

import json

from .bases_meta import DumpMeta
from .class_helper import _META
from .enums import LetterCase
from .lazy_imports import toml, toml_w, yaml
from .loader_selection import asdict, fromdict, fromlist
from .models import Container
from .serial_json import JSONSerializable


class JSONListWizard(JSONSerializable, str=False):
    """
    A Mixin class that extends :class:`JSONSerializable` (JSONWizard)
    to return :class:`Container` - instead of `list` - objects.

    Note that `Container` objects are simply convenience wrappers around a
    collection of dataclass instances. For all intents and purposes, they
    behave exactly the same as `list` objects, with some added helper methods:

        * ``prettify`` - Convert the list of instances to a *prettified* JSON
          string.

        * ``to_json`` - Convert the list of instances to a JSON string.

        * ``to_json_file`` - Serialize the list of instances and write it to a
          JSON file.

    """
    @classmethod
    def from_json(cls, string, *,
                  decoder=json.loads,
                  **decoder_kwargs):
        """
        Converts a JSON `string` to an instance of the dataclass, or a
        Container (list) of the dataclass instances.
        """
        o = decoder(string, **decoder_kwargs)

        if isinstance(o, dict):
            return fromdict(cls, o)

        return Container[cls](fromlist(cls, o))

    @classmethod
    def from_list(cls, o):
        """
        Converts a Python `list` object to a Container (list) of the dataclass
        instances.
        """
        return Container[cls](fromlist(cls, o))


class JSONFileWizard:
    """
    A Mixin class that makes it easier to interact with JSON files.

    This can be paired with the :class:`JSONSerializable` (JSONWizard) Mixin
    class for more complete extensibility.

    """
    @classmethod
    def from_json_file(cls, file, *,
                       decoder=json.load,
                       **decoder_kwargs):
        """
        Reads in the JSON file contents and converts to an instance of the
        dataclass, or a list of the dataclass instances.
        """
        with open(file) as in_file:
            o = decoder(in_file, **decoder_kwargs)

        return fromdict(cls, o) if isinstance(o, dict) else fromlist(cls, o)

    def to_json_file(self, file, mode='w',
                     encoder=json.dump,
                     **encoder_kwargs):
        """
        Serializes the instance and writes it to a JSON file.
        """
        with open(file, mode) as out_file:
            encoder(asdict(self), out_file, **encoder_kwargs)


class TOMLWizard:
    # noinspection PyUnresolvedReferences
    """
    A Mixin class that makes it easier to interact with TOML data.

    .. NOTE::
      By default, *NO* key transform is used in the TOML dump process.
      In practice, this means that a `snake_case` field name in Python is saved
      as `snake_case` to TOML; however, this can easily be customized without
      the need to sub-class from :class:`JSONWizard`.

    For example:

        >>> @dataclass
        >>> class MyClass(TOMLWizard, key_transform='CAMEL'):
        >>>     ...

    """
    def __init_subclass__(cls, key_transform=LetterCase.NONE):
        """Allow easy setup of common config, such as key casing transform."""
        # Only add the key transform if Meta config has not been specified
        # for the dataclass.
        if key_transform and cls not in _META:
            DumpMeta(key_transform=key_transform).bind_to(cls)

    @classmethod
    def from_toml(cls,
                  string_or_stream, *,
                  decoder=None,
                  header='items',
                  parse_float=float):
        """
        Converts a TOML `string` to an instance of the dataclass, or a list of
        the dataclass instances.

        If ``header`` is provided and the corresponding value in the parsed
        data is a ``list``, the return type is ``List[T]``.
        """
        if decoder is None:  # pragma: no cover
            decoder = toml.loads

        o = decoder(string_or_stream, parse_float=parse_float)

        return (fromlist(cls, maybe_l)
                if (maybe_l := o.get(header)) and isinstance(maybe_l, list)
                else fromdict(cls, o))

    @classmethod
    def from_toml_file(cls, file, *,
                       decoder=None,
                       header='items',
                       parse_float=float):
        """
        Reads the contents of a TOML file and converts them
        into an instance (or list of instances) of the dataclass.

        Similar to :meth:`from_toml`, it can return a list if ``header``
        is specified and points to a list in the TOML data.
        """
        if decoder is None:  # pragma: no cover
            decoder = toml.load

        with open(file, 'rb') as in_file:
            return cls.from_toml(in_file,
                                 decoder=decoder,
                                 header=header,
                                 parse_float=parse_float)

    def to_toml(self,
                /,
                *encoder_args,
                encoder=None,
                multiline_strings=False,
                indent=4):
        """
        Converts a dataclass instance to a TOML `string`.

        Optional parameters include ``multiline_strings``
        for enabling/disabling multiline formatting of strings,
        and ``indent`` for setting the indentation level.
        """
        if encoder is None:  # pragma: no cover
            encoder = toml_w.dumps

        return encoder(asdict(self), *encoder_args,
                       multiline_strings=multiline_strings,
                       indent=indent)

    def to_toml_file(self, file, mode='wb',
                     encoder=None,
                     multiline_strings=False,
                     indent=4):
        """
        Serializes a dataclass instance and writes it to a TOML file.

        By default, opens the file in "write binary" mode.
        """
        if encoder is None:  # pragma: no cover
            encoder = toml_w.dump

        with open(file, mode) as out_file:
            self.to_toml(out_file, encoder=encoder,
                         multiline_strings=multiline_strings,
                         indent=indent)

    @classmethod
    def list_to_toml(cls,
                     instances,
                     header='items',
                     encoder=None,
                     **encoder_kwargs):
        """
        Serializes a ``list`` of dataclass instances into a TOML `string`,
        grouped under a specified header.
        """
        if encoder is None:
            encoder = toml_w.dumps

        list_of_dict = [asdict(o, cls=cls) for o in instances]

        return encoder({header: list_of_dict}, **encoder_kwargs)


class YAMLWizard:
    # noinspection PyUnresolvedReferences
    """
    A Mixin class that makes it easier to interact with YAML data.

    .. NOTE::
      The default key transform used in the YAML dump process is `lisp-case`,
      however this can easily be customized without the need to sub-class
      from :class:`JSONWizard`.

    For example:

        >>> @dataclass
        >>> class MyClass(YAMLWizard, key_transform='CAMEL'):
        >>>     ...

    """
    def __init_subclass__(cls, key_transform=LetterCase.LISP):
        """Allow easy setup of common config, such as key casing transform."""
        # Only add the key transform if Meta config has not been specified
        # for the dataclass.
        if key_transform and cls not in _META:
            DumpMeta(key_transform=key_transform).bind_to(cls)

    @classmethod
    def from_yaml(cls,
                  string_or_stream, *,
                  decoder=None,
                  **decoder_kwargs):
        """
        Converts a YAML `string` to an instance of the dataclass, or a list of
        the dataclass instances.
        """
        if decoder is None:
            decoder = yaml.safe_load

        o = decoder(string_or_stream, **decoder_kwargs)

        return fromdict(cls, o) if isinstance(o, dict) else fromlist(cls, o)

    @classmethod
    def from_yaml_file(cls, file, *,
                       decoder=None,
                       **decoder_kwargs):
        """
        Reads in the YAML file contents and converts to an instance of the
        dataclass, or a list of the dataclass instances.
        """
        with open(file) as in_file:
            return cls.from_yaml(in_file, decoder=decoder,
                                 **decoder_kwargs)

    def to_yaml(self, *,
                encoder=None,
                **encoder_kwargs):
        """
        Converts the dataclass instance to a YAML `string` representation.
        """
        if encoder is None:
            encoder = yaml.dump

        return encoder(asdict(self), **encoder_kwargs)

    def to_yaml_file(self, file, mode='w',
                     encoder = None,
                     **encoder_kwargs):
        """
        Serializes the instance and writes it to a YAML file.
        """
        with open(file, mode) as out_file:
            self.to_yaml(stream=out_file, encoder=encoder,
                         **encoder_kwargs)

    @classmethod
    def list_to_yaml(cls,
                     instances,
                     encoder = None,
                     **encoder_kwargs):
        """
        Converts a ``list`` of dataclass instances to a YAML `string`
        representation.
        """
        if encoder is None:
            encoder = yaml.dump

        list_of_dict = [asdict(o, cls=cls) for o in instances]

        return encoder(list_of_dict, **encoder_kwargs)



import logging
import os
import tempfile
from dataclasses import field, dataclass
from datetime import datetime, time, date, timezone
from pathlib import Path
from textwrap import dedent
from typing import ClassVar, List, Dict, Union, DefaultDict, Set

import pytest

from dataclass_wizard.v0 import EnvWizard, env_field
from dataclass_wizard.v0.errors import MissingVars, ParseError, ExtraData
import dataclass_wizard.v0.bases_meta

from tests._typing import *


log = logging.getLogger(__name__)

# quick access to the `tests/unit` directory
here = Path(__file__).parent


def test_load_and_dump():
    """Basic example with simple types (str, int) and collection types such as list."""

    os.environ.update({
        'hello_world': 'Test',
        'MyStr': 'This STRING',
        'MY_TEST_VALUE123': '11',
        'THIS_Num': '23',
        'my_list': '["1", 2, "3", "4.5", 5.7]',
        'my_other_list': 'rob@test.org, this@email.com , hello-world_123@tst.org,z@ab.c'
    })

    class MyClass(EnvWizard, reload_env=True):
        # these are class-level fields, and should be ignored
        my_cls_var: ClassVar[str]
        other_var = 21

        my_str: str
        this_num: int
        my_list: List[int]
        my_other_list: List[str]
        my_test_value123: int = 21
        # missing from environment
        my_field_not_in_env: str = 'testing'

    e = MyClass()
    log.debug(e.dict())

    assert not hasattr(e, 'my_cls_var')
    assert e.other_var == 21

    assert e.my_str == 'This STRING'
    assert e.this_num == 23
    assert e.my_list == [1, 2, 3, 4, 6]
    assert e.my_other_list == ['rob@test.org', 'this@email.com', 'hello-world_123@tst.org', 'z@ab.c']
    assert e.my_test_value123 == 11
    assert e.my_field_not_in_env == 'testing'

    assert e.to_dict() == {
        'my_str': 'This STRING',
        'this_num': 23,
        'my_list': [1, 2, 3, 4, 6],
        'my_other_list': ['rob@test.org',
                          'this@email.com',
                          'hello-world_123@tst.org',
                          'z@ab.c'],
        'my_test_value123': 11,
        'my_field_not_in_env': 'testing',
    }


def test_load_and_dump_with_dict():
    """Example with more complex types such as dict, TypedDict, and defaultdict."""

    os.environ.update({
        'MY_DICT': '{"123": "True", "5": "false"}',
        'My.Other.Dict': 'some_key=value,  anotherKey=123 ,LastKey=just a test~',
        'My_Default_Dict': '  {  "1.2": "2021-01-02T13:57:21"  }  ',
        'myTypedDict': 'my_bool=true'
    })

    class MyTypedDict(TypedDict):
        my_bool: bool

    # Fix so the forward reference works
    globals().update(locals())

    class ClassWithDict(EnvWizard, reload_env=True):
        class _(EnvWizard.Meta):
            field_to_env_var = {'my_other_dict': 'My.Other.Dict'}

        my_dict: Dict[int, bool]
        my_other_dict: Dict[str, Union[int, str]]
        my_default_dict: DefaultDict[float, datetime]
        my_typed_dict: MyTypedDict

    c = ClassWithDict()
    log.debug(c.dict())

    assert c.my_dict == {123: True, 5: False}

    # note that the value for 'anotherKey' is a string value ('123') here,
    # but we might want to see if we can update it to a numeric value (123)
    # instead.
    assert c.my_other_dict == {
        'some_key': 'value',
        'anotherKey': '123',
        'LastKey': 'just a test~',
    }

    assert c.my_default_dict == {1.2: datetime(2021, 1, 2, 13, 57, 21)}
    assert c.my_typed_dict == {'my_bool': True}

    assert c.to_dict() == {
        'my_dict': {5: False, 123: True},
        'my_other_dict': {'LastKey': 'just a test~',
                          'anotherKey': '123',
                          'some_key': 'value'},
        'my_default_dict': {1.2: '2021-01-02T13:57:21'},
        'my_typed_dict': {'my_bool': True}
    }


def test_load_and_dump_with_aliases():
    """
    Example with fields that are aliased to differently-named env variables
    in the Environment.
    """

    os.environ.update({
        'hello_world': 'Test',
        'MY_TEST_VALUE123': '11',
        'the_number': '42',
        'my_list': '3, 2,  1,0',
        'My_Other_List': 'rob@test.org, this@email.com , hello-world_123@tst.org,z@ab.c'
    })

    class MyClass(EnvWizard, reload_env=True):
        class _(EnvWizard.Meta):
            field_to_env_var = {
                'answer_to_life': 'the_number',
                'emails': ('EMAILS', 'My_Other_List'),
            }

        my_str: str = env_field(('the_string', 'hello_world'))
        answer_to_life: int
        list_of_nums: List[int] = env_field('my_list')
        emails: List[str]
        # added for code coverage.
        # case where `env_field` is used, but an alas is not defined.
        my_test_value123: int = env_field(..., default=21)

    c = MyClass()
    log.debug(c.dict())

    assert c.my_str == 'Test'
    assert c.answer_to_life == 42
    assert c.list_of_nums == [3, 2, 1, 0]
    assert c.emails == ['rob@test.org', 'this@email.com', 'hello-world_123@tst.org', 'z@ab.c']
    assert c.my_test_value123 == 11

    assert c.to_dict() == {
        'answer_to_life': 42,
        'emails': ['rob@test.org',
                   'this@email.com',
                   'hello-world_123@tst.org',
                   'z@ab.c'],
        'list_of_nums': [3, 2, 1, 0],
        'my_str': 'Test',
        'my_test_value123': 11,
    }


def test_load_with_missing_env_variables():
    """
    Test calling the constructor of an `EnvWizard` subclass when the
    associated vars are missing in the Environment.
    """

    class MyClass(EnvWizard):
        missing_field_1: str
        missing_field_2: datetime
        missing_field_3: Dict[str, int]
        default_field: Set[str] = field(default_factory=set)

    with pytest.raises(MissingVars) as e:
        _ = MyClass()

    assert str(e.value) == dedent("""
    `test_load_with_missing_env_variables.<locals>.MyClass` has 3 required fields missing in the environment:
        - missing_field_1 -> missing_field_1
        - missing_field_2 -> missing_field_2
        - missing_field_3 -> missing_field_3

    **Resolution options**

    1. Set a default value for the field:

        class test_load_with_missing_env_variables.<locals>.MyClass:
            missing_field_1: str = ''
            missing_field_2: datetime = None
            missing_field_3: typing.Dict[str, int] = None

    2. Provide the value during initialization:

        instance = test_load_with_missing_env_variables.<locals>.MyClass(missing_field_1='', missing_field_2=None, missing_field_3=None)
    """.rstrip())

    # added for code coverage.
    # test when only missing a single (1) required field.
    with pytest.raises(MissingVars) as e:
        _ = MyClass(missing_field_1='test', missing_field_3='key=123')

    error_info = str(e.value)
    assert '1 required field' in error_info
    assert 'missing_field_2' in error_info


def test_load_with_parse_error():
    os.environ.update(MY_STR='abc')

    class MyClass(EnvWizard, reload_env=True):
        class _(EnvWizard.Meta):
            debug_enabled = True

        my_str: int

    with pytest.raises(ParseError) as e:
        _ = MyClass()

    assert str(e.value.base_error) == "invalid literal for int() with base 10: 'abc'"
    assert e.value.kwargs['env_variable'] == 'MY_STR'

    del os.environ['MY_STR']


def test_load_with_parse_error_when_env_var_is_specified():
    """
    Raising `ParseError` when a dataclass field to env var mapping is
    specified. Added for code coverage.
    """

    os.environ.update(MY_STR='abc')

    class MyClass(EnvWizard, reload_env=True):
        class _(EnvWizard.Meta):
            debug_enabled = True

        a_string: int = env_field('MY_STR')

    with pytest.raises(ParseError) as e:
        _ = MyClass()

    assert str(e.value.base_error) == "invalid literal for int() with base 10: 'abc'"
    assert e.value.kwargs['env_variable'] == 'MY_STR'

    del os.environ['MY_STR']


def test_load_with_dotenv_file():
    """Test reading from the `.env` file in project root directory."""

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            env_file = True

        my_str: int
        my_time: time
        my_date: date = None

    assert MyClass().dict() == {'my_str': 42,
                                'my_time': time(15, 20),
                                'my_date': date(2022, 1, 21)}


def test_load_with_dotenv_file_with_path():
    """Test reading from the `.env.test` file in `tests/unit` directory."""

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            env_file = here / '.env.test'
            key_lookup_with_load = 'PASCAL'

        my_value: float
        my_dt: datetime
        another_date: date

    c = MyClass()

    assert c.dict() == {'my_value': 1.23,
                        'my_dt': datetime(2022, 4, 27, 16, 30, 45, tzinfo=timezone.utc),
                        'another_date': date(2021, 12, 17)}

    expected_json = '{"another_date": "2021-12-17", "my_dt": "2022-04-27T16:30:45Z", "my_value": 1.23}'
    assert c.to_json(sort_keys=True) == expected_json


def test_load_with_tuple_of_dotenv_and_env_file_param_to_init():
    """
    Test when `env_file` is specified as a tuple of dotenv files, and
    the `_env_file` parameter is also passed in to the constructor
    or __init__() method.
    """

    os.environ.update(
        MY_STR='default from env',
        myValue='3322.11',
        Other_Key='5',
    )

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            env_file = '.env', here / '.env.test'
            key_lookup_with_load = 'PASCAL'

        my_value: float
        my_str: str
        other_key: int = 3

    # pass `_env_file=False` so we don't load the Meta `env_file`
    c = MyClass(_env_file=False, _reload=True)

    assert c.dict() == {'my_str': 'default from env',
                        'my_value': 3322.11,
                        'other_key': 5}

    # load variables from the Meta `env_file` tuple, and also pass
    # in `other_key` to the constructor method.
    c = MyClass(other_key=7)

    assert c.dict() == {'my_str': '42',
                        'my_value': 1.23,
                        'other_key': 7}

    # load variables from the `_env_file` argument to the constructor
    # method, overriding values from `env_file` in the Meta config.
    c = MyClass(_env_file=here / '.env.prod')

    assert c.dict() == {'my_str': 'hello world!',
                        'my_value': 3.21,
                        'other_key': 5}

    del os.environ['MY_STR']


def test_load_when_constructor_kwargs_are_passed():
    """
    Using the constructor method of an `EnvWizard` subclass when
    passing keyword arguments instead of the Environment.
    """
    os.environ.update(MY_STRING_VAR='hello world')

    class MyTestClass(EnvWizard, reload_env=True):
        my_string_var: str

    c = MyTestClass(my_string_var='test!!')
    assert c.my_string_var == 'test!!'

    c = MyTestClass()
    assert c.my_string_var == 'hello world'

# TODO

# def test_extra_keyword_arguments_when_deny_extra():
#     """
#     Passing extra keyword arguments to the constructor method of an `EnvWizard`
#     subclass raises an error by default, as `Extra.DENY` is the default behavior.
#     """
#
#     os.environ['A_FIELD'] = 'hello world!'
#
#     class MyClass(EnvWizard, reload_env=True):
#         a_field: str
#
#     with pytest.raises(ExtraData) as e:
#         _ = MyClass(another_field=123, third_field=None)
#
#     log.error(e.value)
#
#
# def test_extra_keyword_arguments_when_allow_extra():
#     """
#     Passing extra keyword arguments to the constructor method of an `EnvWizard`
#     subclass does not raise an error and instead accepts or "passes through"
#     extra keyword arguments, when `Extra.ALLOW` is specified for the
#     `extra` Meta field.
#     """
#
#     os.environ['A_FIELD'] = 'hello world!'
#
#     class MyClass(EnvWizard, reload_env=True):
#
#         class _(EnvWizard.Meta):
#             extra = 'ALLOW'
#
#         a_field: str
#
#     c = MyClass(another_field=123, third_field=None)
#
#     assert getattr(c, 'another_field') == 123
#     assert hasattr(c, 'third_field')
#
#     assert c.to_json() == '{"a_field": "hello world!"}'
#
#
# def test_extra_keyword_arguments_when_ignore_extra():
#     """
#     Passing extra keyword arguments to the constructor method of an `EnvWizard`
#     subclass does not raise an error and instead ignores extra keyword
#     arguments, when `Extra.IGNORE` is specified for the `extra` Meta field.
#     """
#
#     os.environ['A_FIELD'] = 'hello world!'
#
#     class MyClass(EnvWizard, reload_env=True):
#
#         class _(EnvWizard.Meta):
#             extra = 'IGNORE'
#
#         a_field: str
#
#     c = MyClass(another_field=123, third_field=None)
#
#     assert not hasattr(c, 'another_field')
#     assert not hasattr(c, 'third_field')
#
#     assert c.to_json() == '{"a_field": "hello world!"}'


def test_init_method_declaration_is_logged_when_debug_mode_is_enabled(mock_debug_log):

    class _EnvSettings(EnvWizard):

        class _(EnvWizard.Meta):
            debug_enabled = True
            extra = 'ALLOW'

        auth_key: str = env_field('my_auth_key')
        api_key: str = env_field(('hello', 'test'))
        domains: Set[str] = field(default_factory=set)
        answer_to_life: int = 42

    # assert that the __init__() method declaration is logged
    assert mock_debug_log.records[-1].levelname == 'DEBUG'
    assert 'Generated function code' in mock_debug_log.records[-3].message

    # reset global flag for other tests that
    # rely on `debug_enabled` functionality
    dataclass_wizard.v0.bases_meta._debug_was_enabled = False


def test_load_with_tuple_of_dotenv_and_env_prefix_param_to_init():
    """
    Test when `env_file` is specified as a tuple of dotenv files, and
    the `_env_file` parameter is also passed in to the constructor
    or __init__() method. Additionally, test prefixing environment
    variables using `Meta.env_prefix` and `_env_prefix` in __init__().
    """

    os.environ.update(
        PREFIXED_MY_STR='prefixed string',
        PREFIXED_MY_VALUE='12.34',
        PREFIXED_OTHER_KEY='10',
        MY_STR='default from env',
        MY_VALUE='3322.11',
        OTHER_KEY='5',
    )

    class MyClass(EnvWizard):
        class _(EnvWizard.Meta):
            env_file = '.env', here / '.env.test'
            env_prefix = 'PREFIXED_'  # Static prefix
            key_lookup_with_load = 'PASCAL'

        my_value: float
        my_str: str
        other_key: int = 3

    # Test without prefix
    c = MyClass(_env_file=False, _reload=True,
                _env_prefix=None)

    assert c.dict() == {'my_str': 'default from env',
                        'my_value': 3322.11,
                        'other_key': 5}

    # Test with Meta.env_prefix applied
    c = MyClass(other_key=7)

    assert c.dict() == {'my_str': 'prefixed string',
                        'my_value': 12.34,
                        'other_key': 7}

    # Override prefix dynamically with _env_prefix
    c = MyClass(_env_file=False, _env_prefix='', _reload=True)

    assert c.dict() == {'my_str': 'default from env',
                        'my_value': 3322.11,
                        'other_key': 5}

    # Dynamically set a new prefix via _env_prefix
    c = MyClass(_env_prefix='PREFIXED_')

    assert c.dict() == {'my_str': 'prefixed string',
                        'my_value': 12.34,
                        'other_key': 10}

    # Otherwise, this would take priority, as it's named `My_Value` in `.env.prod`
    del os.environ['MY_VALUE']

    # Load from `_env_file` argument, ignoring prefixes
    c = MyClass(_reload=True, _env_file=here / '.env.prod', _env_prefix='')

    assert c.dict() == {'my_str': 'hello world!',
                        'my_value': 3.21,
                        'other_key': 5}

    del os.environ['MY_STR']


def test_env_prefix_with_env_file():
    """
    Test `env_prefix` with `env_file` where file has prefixed env variables.

    Contents of `.env.prefix`:
        MY_PREFIX_STR='my prefix value'
        MY_PREFIX_BOOL=t
        MY_PREFIX_INT='123.0'

    """
    class MyPrefixTest(EnvWizard):
        class _(EnvWizard.Meta):
            env_prefix = 'MY_PREFIX_'
            env_file = here / '.env.prefix'

        str: str
        bool: bool
        int: int

    expected = MyPrefixTest(str='my prefix value',
                            bool=True,
                            int=123)

    assert MyPrefixTest() == expected


def test_secrets_dir_and_override():
    """
    Test `Meta.secrets_dir` and `_secrets_dir` for handling secrets.
    """
    # Create temporary directories and files to simulate secrets
    with tempfile.TemporaryDirectory() as default_secrets_dir, tempfile.TemporaryDirectory() as override_secrets_dir:
        # Paths for default secrets
        default_dir_path = Path(default_secrets_dir)
        (default_dir_path / "MY_SECRET_KEY").write_text("default-secret-key")
        (default_dir_path / "ANOTHER_SECRET").write_text("default-another-secret")

        # Paths for override secrets
        override_dir_path = Path(override_secrets_dir)
        (override_dir_path / "MY_SECRET_KEY").write_text("override-secret-key")
        (override_dir_path / "NEW_SECRET").write_text("new-secret-value")

        # Define an EnvWizard class with Meta.secrets_dir
        class MySecretClass(EnvWizard):
            class _(EnvWizard.Meta):
                secrets_dir = default_dir_path  # Static default secrets directory

            my_secret_key: str
            another_secret: str = "default"
            new_secret: str = "default-new"

        # Test case 1: Use Meta.secrets_dir
        instance = MySecretClass()
        assert instance.dict() == {
            "my_secret_key": "default-secret-key",
            "another_secret": "default-another-secret",
            "new_secret": "default-new",
        }

        # Test case 2: Override secrets_dir using _secrets_dir
        instance = MySecretClass(_secrets_dir=override_dir_path)
        assert instance.dict() == {
            "my_secret_key": "override-secret-key",  # Overridden by override directory
            "another_secret": "default-another-secret",  # Still from Meta.secrets_dir
            "new_secret": "new-secret-value",  # Only in override directory
        }

        # Test case 3: Missing secrets fallback to defaults
        instance = MySecretClass(_reload=True)
        assert instance.dict() == {
            "my_secret_key": "default-secret-key",  # From default directory
            "another_secret": "default-another-secret",  # From default directory
            "new_secret": "default-new",  # From the field default
        }

        # Test case 4: Invalid secrets_dir scenarios
        # Case 4a: Directory doesn't exist (ignored with warning)
        instance = MySecretClass(_secrets_dir=(default_dir_path, Path("/non/existent/directory")),
                                 _reload=True)
        assert instance.dict() == {
            "my_secret_key": "default-secret-key",  # Fallback to default secrets
            "another_secret": "default-another-secret",
            "new_secret": "default-new",
        }

        # Case 4b: secrets_dir is a file (raises error)
        with tempfile.NamedTemporaryFile() as temp_file:
            invalid_secrets_path = Path(temp_file.name)
            with pytest.raises(ValueError, match="Secrets directory .* is a file, not a directory"):
                MySecretClass(_secrets_dir=invalid_secrets_path, _reload=True)


def test_env_wizard_handles_nested_dataclass_field_with_multiple_input_types():
    """
    Test that EnvWizard correctly handles a field typed as a nested dataclass:

    - When specified as an environment variable (JSON-encoded string).
    - When passed as a dictionary to the constructor.
    - When passed as an instance of the nested dataclass.
    """

    @dataclass
    class DatabaseSettings:
        host: str
        port: int

    class Settings(EnvWizard):
        database: DatabaseSettings

        class Config(EnvWizard.Meta):
            env_prefix='test'
            env_nested_delimiter = '_'

    # Field `database` is specified as an env var
    os.environ['testdatabase'] = '{"host": "localhost", "port": "5432"}'

    # need to `_reload` due to other test cases
    settings = Settings(_reload=True)
    assert settings == Settings(database=DatabaseSettings(host='localhost', port=5432))

    # Field `database` is specified as a dict
    settings = Settings(database={"host": "localhost", "port": "4000"})
    assert settings == Settings(database=DatabaseSettings(host='localhost', port=4000))

    # Field `database` is passed in to constructor (__init__)
    settings = Settings(database=(db := DatabaseSettings(host='localhost', port=27017)))
    assert settings.database == db



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Collection, Iterable, Sequence
from dataclasses import MISSING, Field, is_dataclass
from datetime import date, datetime, time
from enum import Enum
from json import JSONEncoder, dumps
from typing import Any, Callable, ClassVar
from uuid import UUID

from .utils._string_conv import normalize

# added as we can't import from `type_def`, as we run into a circular import.
JSONObject = dict[str, Any]

_SafeEncoder = None


def type_name(obj: type) -> str:
    """Return the type or class name of an object"""
    from .utils._typing_compat import is_generic

    # for type generics like `dict[str, float]`, we want to return
    # the subscripted value as is, rather than simply accessing the
    # `__name__` property, which in this case would be `dict` instead.
    if is_generic(obj):
        return str(obj)

    return getattr(obj, '__qualname__', getattr(obj, '__name__', repr(obj)))


def show_deprecation_warning(
    fn: Callable | str,
    reason: str,
    fmt: str = "Deprecated function {name} ({reason})."
) -> None:
    """
    Display a deprecation warning for a given function.

    @param fn: Function which is deprecated.
    @param reason: Reason for the deprecation.
    @param fmt: Format string for the name/reason.
    """
    import warnings
    warnings.simplefilter('always', DeprecationWarning)
    warnings.warn(
        fmt.format(name=getattr(fn, '__name__', fn), reason=reason),
        category=DeprecationWarning,
        stacklevel=2,
    )


def _get_safe_encoder() -> type[JSONEncoder]:
    from ._dumpers import asdict

    global _SafeEncoder
    if _SafeEncoder is not None:
        return _SafeEncoder

    class _LocalSafeEncoder(JSONEncoder):
        def default(self, o: Any) -> Any:
            if is_dataclass(o):
                return asdict(o)
            if isinstance(o, Enum):
                return o.value
            if isinstance(o, UUID):
                return o.hex
            if isinstance(o, (datetime, time)):
                return o.isoformat().replace('+00:00', 'Z', 1)
            if isinstance(o, date):
                return o.isoformat()
            return str(o)

    _SafeEncoder = _LocalSafeEncoder
    return _SafeEncoder


def safe_dumps(o: Any, **kwargs: Any) -> str:
    # never let callers override cls; this is for errors, not a general API
    try:
        return dumps(o, cls=_get_safe_encoder(), **kwargs)
    except TypeError:
        # returning `o` here is inconsistent; callers expect str
        return str(o)


class JSONWizardError(ABC, Exception):
    """
    Base error class, for errors raised by this library.
    """

    _TEMPLATE: ClassVar[str]

    @property
    def class_name(self) -> str | None:
        return self._class_name or self._default_class_name

    @class_name.setter
    def class_name(self, cls: str | type | None) -> None:
        # Set parent class for errors
        self.parent_cls = cls
        # Set class name
        if getattr(self, '_class_name', None) is None:
            # noinspection PyAttributeOutsideInit
            self._class_name = self.name(cls)

    @property
    def parent_cls(self) -> type | None:
        return self._parent_cls

    @parent_cls.setter
    def parent_cls(self, cls: type | None):
        # noinspection PyAttributeOutsideInit
        self._parent_cls = cls

    @staticmethod
    def name(obj) -> str:
        """Return the type or class name of an object"""
        # Uses short-circuiting with `or` to efficiently
        # return the first valid name.
        return (getattr(obj, '__qualname__', None)
                or getattr(obj, '__name__', None)
                or str(obj))

    @property
    @abstractmethod
    def message(self) -> str:
        """
        Format and return an error message.
        """

    def __str__(self):
        return self.message


class ParseError(JSONWizardError):
    """
    Base error when an error occurs during the JSON load process.
    """

    _TEMPLATE = ('Failed to {p} field `{field}` in class `{cls}`.{expectation}\n'
                 '  phase: {p}\n'
                 '{value}'
                 '  error: {e!s}')

    def __init__(self, base_err: Exception,
                 obj: Any,
                 ann_type: type | Iterable | None,
                 phase: str,
                 _default_class: type | None = None,
                 _field_name: str | None = None,
                 _json_object: Any = None,
                 **kwargs):

        super().__init__()

        self.phase = phase
        self.obj = obj
        self.obj_type = type(obj)
        self.ann_type = ann_type
        self.base_error = base_err
        self.kwargs = kwargs
        self._class_name = None
        self._default_class_name = self.name(_default_class) \
            if _default_class else None
        self._field_name = _field_name
        self._json_object = _json_object
        self.fields = None

    @property
    def field_name(self) -> str | None:
        return self._field_name

    @field_name.setter
    def field_name(self, name: str | None):
        if self._field_name is None:
            self._field_name = name

    @property
    def json_object(self):
        return self._json_object

    @json_object.setter
    def json_object(self, json_obj):
        if self._json_object is None:
            self._json_object = json_obj

    @property
    def message(self) -> str:
        if self.obj_type is type:
            obj_type = self.name(self.obj)
            expectation = ''
            value = f'  value_type: {obj_type}\n'
        else:
            obj_type = self.name(self.obj_type)
            expectation = f' Expected a type {self.ann_type}, got {obj_type}.'
            value = f'  value: {self.obj!r}\n'

        ann_type = self.name(
            self.ann_type if self.ann_type is not None
            else next((f.type for f in self.fields
                       if f.name == self._field_name), None))

        msg = self._TEMPLATE.format(
            expectation=expectation,
            cls=self.class_name, field=self.field_name,
            e=self.base_error, value=value, p=self.phase,
            ann_type=ann_type)

        if self.json_object:
            self.kwargs['json_object'] = safe_dumps(self.json_object)

        if self.kwargs:
            sep = '\n  '
            parts = sep.join(f'{k}: {v if isinstance(v, str) else repr(v)}'
                             for k, v in self.kwargs.items())
            msg = f'{msg}{sep}{parts}'

        return msg


class ExtraData(JSONWizardError):
    """
    Error raised when extra keyword arguments are passed in to the constructor
    or `__init__()` method of an `EnvWizard` subclass.

    Note that this error class is raised by default, unless a value for the
    `extra` field is specified in the :class:`Meta` class.
    """

    _TEMPLATE = ('{cls}.__init__() received extra keyword arguments:\n'
                 '  extras: {extra_kwargs!r}\n'
                 '  fields: {field_names!r}\n'
                 '  resolution: specify a value for `extra` in the Meta '
                 'config for the class, to control how extra keyword '
                 'arguments are handled.')

    def __init__(self,
                 cls: type,
                 extra_kwargs: Collection[str],
                 field_names: Collection[str]):

        super().__init__()

        self.class_name = type_name(cls)
        self.extra_kwargs = extra_kwargs
        self.field_names = field_names

    @property
    def message(self) -> str:
        msg = self._TEMPLATE.format(
            cls=self.class_name,
            extra_kwargs=self.extra_kwargs,
            field_names=self.field_names,
        )

        return msg


class MissingFields(JSONWizardError):
    """
    Error raised when unable to create a class instance (most likely due to
    missing arguments)
    """

    _TEMPLATE = ('`{cls}.__init__()` missing required fields.\n'
                 '  Provided: {fields!r}\n'
                 '  Missing: {missing_fields!r}\n'
                 '{expected_keys}'
                 '  Input JSON: {json_string}'
                 '{e}')

    def __init__(self, base_err: Exception | None,
                 obj: JSONObject,
                 cls: type,
                 cls_fields: tuple[Field, ...],
                 cls_kwargs: JSONObject | None = None,
                 missing_fields: Collection[str] | None = None,
                 missing_keys: Collection[str] | None = None,
                 **kwargs):

        super().__init__()

        self.obj = obj

        if missing_fields:
            self.fields = [f.name for f in cls_fields
                           if f.name not in missing_fields
                           and f.default is MISSING
                           and f.default_factory is MISSING]
            self.missing_fields = missing_fields
        else:
            self.fields = list(cls_kwargs.keys())
            self.missing_fields = [f.name for f in cls_fields
                                   if f.name not in self.fields
                                   and f.default is MISSING
                                   and f.default_factory is MISSING]

        self.base_error = base_err
        self.missing_keys = missing_keys
        self.kwargs = kwargs
        self.class_name: str = self.name(cls)
        self.parent_cls = cls
        self.all_fields = cls_fields

    @property
    def message(self) -> str:
        if isinstance(self.obj, list):
            keys = [f.name for f in self.all_fields]
            obj = dict(zip(keys, self.obj))
        else:
            obj = self.obj

        # check if any field names match, and where the key transform could be the cause
        # see https://github.com/rnag/dataclass-wizard/issues/54 for more info

        normalized_json_keys = [normalize(key) for key in obj]
        if (is_dataclass(self.parent_cls) and
            next((f for f in self.missing_fields if normalize(f) in normalized_json_keys), None)):
            from ._loaders import get_loader
            from .enums import KeyCase

            key_transform = get_loader(self.parent_cls).transform_json_field
            if isinstance(key_transform, KeyCase):
                if key_transform.value is None:
                    key_transform = f'{key_transform.name}'
                else:
                    key_transform = f'{key_transform.value.f.__name__}()'
            elif key_transform is not None:
                key_transform = f'{getattr(key_transform, "__name__", key_transform)}()'

            self.kwargs['Key Transform'] = key_transform
            self.kwargs['Resolution'] = 'For more details, please see https://github.com/rnag/dataclass-wizard/issues/54'

        self.kwargs['Resolution'] = ('Ensure that all required fields are provided in the input. '
                                     'For more details, see:\n'
                                     '    https://github.com/rnag/dataclass-wizard/discussions/167')

        if self.base_error is not None:
            e = f'\n  error: {self.base_error!s}'
        else:
            e = ''

        if self.missing_keys is not None:
            expected_keys = f'  Expected Keys: {self.missing_keys!r}\n'
        else:
            expected_keys = ''

        msg = self._TEMPLATE.format(
            cls=self.class_name,
            json_string=safe_dumps(self.obj),
            e=e,
            fields=self.fields,
            expected_keys=expected_keys,
            missing_fields=self.missing_fields)

        if self.kwargs:
            sep = '\n  '
            parts = sep.join(f'{k}: {v if isinstance(v, str) else repr(v)}'
                             for k, v in self.kwargs.items())
            msg = f'{msg}{sep}{parts}'

        return msg


class UnknownKeysError(JSONWizardError):
    """
    Error raised when unknown JSON key(s) are
    encountered in the JSON load process.

    Note that this error class is only raised when the
    `on_unknown_key='RAISE'` flag is enabled in
    the :class:`Meta` class.
    """

    _TEMPLATE = ('One or more JSON keys are not mapped to the dataclass schema for class `{cls}`.\n'
                 '  Unknown key{s}: {unknown_keys!r}\n'
                 '  Dataclass fields: {fields!r}\n'
                 '  Input JSON object: {json_string}')

    def __init__(self,
                 unknown_keys: list[str] | str,
                 obj: JSONObject,
                 cls: type,
                 cls_fields: tuple[Field, ...], **kwargs):
        super().__init__()

        self.unknown_keys = unknown_keys
        self.obj = obj
        self.fields = [f.name for f in cls_fields]
        self.kwargs = kwargs
        self.class_name: str = self.name(cls)

    @property
    def json_key(self):
        show_deprecation_warning(
            UnknownKeysError.json_key.fget,
            'use `unknown_keys` instead',
        )
        return self.unknown_keys

    @property
    def message(self) -> str:
        if not isinstance(self.unknown_keys, str) and len(self.unknown_keys) > 1:
            s = 's'
        else:
            s = ''

        msg = self._TEMPLATE.format(
            cls=self.class_name,
            s=s,
            json_string=safe_dumps(self.obj),
            fields=self.fields,
            unknown_keys=self.unknown_keys)

        if self.kwargs:
            sep = '\n  '
            parts = sep.join(f'{k}: {v if isinstance(v, str) else repr(v)}'
                             for k, v in self.kwargs.items())
            msg = f'{msg}{sep}{parts}'

        return msg


# Alias for backwards-compatibility.
UnknownJSONKey = UnknownKeysError


class MissingData(ParseError):
    """
    Error raised when unable to create a class instance, as the JSON object
    is None.
    """

    _TEMPLATE = ('Failure loading class `{cls}`. '
                 'Missing value for field (expected a dict, got None)\n'
                 '  dataclass field: {field!r}\n'
                 '  resolution: annotate the field as '
                 '`Optional[{nested_cls}]` or `{nested_cls} | None`')

    def __init__(self, nested_cls: type, **kwargs):
        super().__init__(self, None, nested_cls, 'load', **kwargs)
        self.nested_class_name: str = self.name(nested_cls)

        # self.nested_class_name: str = type_name(nested_cls)

    @property
    def message(self) -> str:
        msg = self._TEMPLATE.format(
            cls=self.class_name,
            nested_cls=self.nested_class_name,
            json_string=safe_dumps(self.obj),
            field=self.field_name,
        )

        if self.kwargs:
            sep = '\n  '
            parts = sep.join(f'{k}: {v if isinstance(v, str) else repr(v)}'
                             for k, v in self.kwargs.items())
            msg = f'{msg}{sep}{parts}'

        return msg


class InvalidConditionError(JSONWizardError):
    """
    Error raised when a condition is not wrapped in ``SkipIf``.
    """

    _TEMPLATE = ('Failure parsing annotations for class `{cls}`. '
                 'Field has an invalid condition.\n'
                 '  dataclass field: {field!r}\n'
                 '  resolution: Wrap conditions inside SkipIf().`')

    def __init__(self, cls: type, field_name: str):
        super().__init__()

        self.class_name: str = self.name(cls)
        self.field_name: str = field_name

    @property
    def message(self) -> str:
        return self._TEMPLATE.format(cls=self.class_name,
                                     field=self.field_name)


class MissingVars(JSONWizardError):
    """
    Error raised when unable to create an instance of a EnvWizard subclass
    (most likely due to missing environment variables in the Environment)

    """
    _TEMPLATE = ('\n`{cls}` has {prefix} missing in the environment:\n'
                 '{fields}\n\n'
                 '**Resolution options**\n\n'
                 '1. Set a default value for the field:\n\n'
                 '{def_resolution}'
                 '\n\n'
                 '2. Provide the value during initialization:\n\n'
                 '    {init_resolution}')

    def __init__(self,
                 cls: type,
                 missing_vars: Sequence[tuple[str, str | None, str, Any]]):

        super().__init__()

        indent = ' ' * 4

        #   - `name` (mapped to `CUSTOM_A_NAME`)
        self.class_name: str = type_name(cls)
        self.fields = '\n'.join([f'{indent}- {f[0]} -> {f[1]}' for f in missing_vars])
        self.def_resolution = '\n'.join([f'{indent}class {self.class_name}:'] +
                                        [f'{indent * 2}{f}: {typ} = {default!r}'
                                         for (f, _, typ, default) in missing_vars])

        init_vars = ', '.join([f'{f}={default!r}' for (f, _, typ, default) in missing_vars])
        self.init_resolution = f'instance = {self.class_name}({init_vars})'

        num_fields = len(missing_vars)
        self.prefix = f'{len(missing_vars)} required field{"s" if num_fields > 1 else ""}'

    @property
    def message(self) -> str:
        msg = self._TEMPLATE.format(
            cls=self.class_name,
            prefix=self.prefix,
            fields=self.fields,
            def_resolution=self.def_resolution,
            init_resolution=self.init_resolution,
        )

        return msg



import dataclasses
import types
from typing import Any

from _typeshed import Incomplete

def is_builtin_class(cls: type) -> bool: ...

class FunctionBuilder:
    current_function: Incomplete
    functions: Incomplete
    globals: Incomplete
    indent_level: Incomplete
    namespace: Incomplete
    prev_function: Incomplete
    def __init__(self) -> None:
        ...
    def __ior__(self, other): ...
    def __enter__(self): ...
    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None): ...
    def function(self, name: str, args: list,
                 return_type=dataclasses.MISSING,
                 locals: Incomplete | None = ...) -> FunctionBuilder: ...
    def _with_new_block(self, name: str,
                        condition: str | None = None,
                        comment: Any = '') -> FunctionBuilder: ...
    def for_(self, condition: str) -> FunctionBuilder: ...
    def if_(self, condition: str, comment: Any = ...) -> FunctionBuilder: ...
    def elif_(self, condition: str) -> FunctionBuilder: ...
    def else_(self) -> FunctionBuilder: ...
    def try_(self) -> FunctionBuilder: ...
    def except_(self, cls: type, var_name: str | None = ..., *custom_classes: type): ...
    def except_multi(self, *classes: type): ...
    def break_(self): ...
    def add_line(self, line: str): ...
    def add_lines(self, *lines: str): ...
    def increase_indent(self): ...
    def decrease_indent(self): ...
    def finalize_function(self): ...
    def create_functions(self, _globals: Incomplete | None = ...): ...



from __future__ import annotations

import json
import logging
import os
from collections import ChainMap
from collections.abc import Mapping

# noinspection PyUnresolvedReferences,PyProtectedMember
from dataclasses import (  # type: ignore
    _FIELD_INITVAR,
    _POST_INIT_NAME,
    MISSING,
    Field,
)
from typing import TYPE_CHECKING, Any, Callable

from ._bases import AbstractEnvMeta
from ._bases_meta import BaseEnvWizardMeta, EnvMeta, register_type
from ._class_helper import (
    DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD,
    call_meta_initializer_if_needed,
    resolve_dataclass_field_to_env_for_load,
)
from ._decorators import cached_class_property
from ._dumpers import asdict
from ._loaders import LoadMixin as V1LoadMixin
from ._loaders import get_loader
from ._log import LOG, enable_library_debug_logging
from ._meta_cache import get_meta
from ._models import MAPPING_ORIGINS, SEQUENCE_ORIGINS, Extras, TypeInfo
from ._path_util import get_dotenv_map, get_secrets_map
from ._type_conv import as_dict, as_list
from ._type_def import META, JSONObject, T, dataclass_transform
from .constants import CATCH_ALL, PACKAGE_NAME
from .enums import EnvKeyStrategy, EnvPrecedence
from .errors import (
    JSONWizardError,
    MissingData,
    MissingVars,
    ParseError,
    type_name,
)
from .utils._dataclass_compat import (
    SEEN_DEFAULT,
    apply_env_wizard_dataclass,
    dataclass_field_names,
    dataclass_fields,
    dataclass_init_field_names,
    dataclass_init_fields,
    dataclass_needs_refresh,
    set_new_attribute,
)
from .utils._function_builder import FunctionBuilder
from .utils._object_path import env_safe_get
from .utils._string_conv import possible_env_vars
from .utils._typing_compat import eval_forward_ref_if_needed

if TYPE_CHECKING:
    from ._env import E_, EnvInit


def env_config(**kw):
    return kw


_PRECEDENCE_ORDER: dict[EnvPrecedence, tuple[str, ...]] = {
    EnvPrecedence.SECRETS_ENV_DOTENV: ('secrets', 'env', 'dotenv'),
    EnvPrecedence.SECRETS_DOTENV_ENV: ('secrets', 'dotenv', 'env'),
    EnvPrecedence.ENV_ONLY: ('env', ),
}


def _pre_decoder(_cls: V1LoadMixin, container_tp: type, tp: TypeInfo, extras: Extras):
    if tp.i == 1:  # Outermost container (first seen in field annotation)
        if container_tp in SEQUENCE_ORIGINS:
            tp.ensure_in_locals(extras, as_list=as_list)
            return tp.replace(val_name=f'as_list({tp.v()})')

        elif container_tp in MAPPING_ORIGINS:
            tp.ensure_in_locals(extras, as_dict=as_dict)
            return tp.replace(val_name=f'as_dict({tp.v()})')

    return tp


@dataclass_transform(kw_only_default=True)
class EnvWizard:
    __slots__ = ()

    class Meta(BaseEnvWizardMeta):
        """
        Inner meta class that can be extended by sub-classes for additional
        customization with the environment load process.
        """
        __slots__ = ()

        # Class attribute to enable detection of the class type.
        __is_inner_meta__ = True

        def __init_subclass__(cls):
            # Set the `__init_subclass__` method here, so we can ensure it
            # doesn't run for the `EnvWizard.Meta` class.
            return cls._init_subclass()

    def __init__(self, **kwargs):
        __init_fn__ = load_func_for_dataclass(
            self.__class__,
            LoadMixin,
            AbstractEnvMeta,
        )
        __init_fn__(self, **kwargs)

    def __init_subclass__(cls,
                          debug: bool = False,
                          _apply_dataclass=True,
                          **dc_kwargs):
        super().__init_subclass__()

        # skip classes provided by this library.
        if cls.__module__.startswith(f'{PACKAGE_NAME}.'):  # pragma: no cover
            return

        # Apply the @dataclass decorator.
        if _apply_dataclass and dataclass_needs_refresh(cls):
            # noinspection PyArgumentList
            apply_env_wizard_dataclass(cls, dc_kwargs)

        load_meta_kwargs = {'pre_decoder': _pre_decoder}

        if debug:
            lvl = logging.DEBUG if isinstance(debug, bool) else debug
            enable_library_debug_logging(lvl)
            # set `debug` flag for the class's Meta
            load_meta_kwargs['debug'] = lvl

        EnvMeta(**load_meta_kwargs).bind_to(cls)

        # Calls the Meta initializer when inner :class:`Meta` is sub-classed.
        call_meta_initializer_if_needed(cls)

    _dcw_env_cache_secrets = classmethod(get_secrets_map)
    _dcw_env_cache_dotenv = classmethod(get_dotenv_map)

    __field_names__ = cached_class_property(dataclass_field_names)

    register_type = classmethod(register_type)

    def raw_dict(self: E_) -> JSONObject:
        """
        Same as ``__dict__``, but only returns values for fields defined
        on the `EnvWizard` instance. See :attr:`__field_names__` for more info.

        .. NOTE::
           The values in the returned dictionary object are not needed to be
           JSON serializable. Use :meth:`to_dict` if this is required.
        """

    to_dict = asdict

    def to_json(self, *,
                encoder=json.dumps,
                **encoder_kwargs):

        return encoder(asdict(self), **encoder_kwargs)


def load_func_for_dataclass(
        cls,
        loader_cls=None,
        base_meta_cls=AbstractEnvMeta,
) -> Callable[[T, dict[str, Any]], None] | None:

    # Tuple describing the fields of this dataclass.
    fields = dataclass_fields(cls)

    cls_init_fields = dataclass_init_fields(cls, True)
    cls_init_field_names = dataclass_init_field_names(cls)

    # Does this class have a post-init function?
    has_post_init = hasattr(cls, _POST_INIT_NAME)

    # Get the loader for the class, or create a new one as needed.
    cls_loader = get_loader(cls, base_cls=loader_cls or LoadMixin)

    cls_name = cls.__name__

    fn_name = f'__{PACKAGE_NAME}_init_{cls_name}__'
    raw_dict_name = f'__{PACKAGE_NAME}_raw_dict_{cls_name}__'

    # Get the meta config for the class, or the default config otherwise.
    meta = get_meta(cls, base_meta_cls)

    # if extras is None:  # we are being run for the main dataclass
    # If the `recursive` flag is enabled and a Meta config is provided,
    # apply the Meta recursively to any nested classes.
    #
    # Else, just use the base `AbstractMeta`.
    config: META = meta if meta.recursive else base_meta_cls

    # Initialize the FuncBuilder
    fn_gen = FunctionBuilder()

    new_locals = {
        'cls': cls,
        'fields': fields,
    }

    # noinspection PyTypeChecker
    extras: Extras = {
        'config': config,
        'cls': cls,
        'cls_name': cls_name,
        'locals': new_locals,
        'recursion_guard': {cls: fn_name},
        'fn_gen': fn_gen,
    }

    _globals = {
        'os': os,
        'ChainMap': ChainMap,
        'MISSING': MISSING,
        'ParseError': ParseError,
        'MissingVars': MissingVars,
        'add': _add_missing_var,
        're_raise': re_raise,
    }

    # we are being run for a nested dataclass
    # NOTE: I don't believe this path exists, since `_loaders.from_dict`
    # is used for nested dataclasses.
    #
    # else:
    #     is_main_class = False

    # default `load_case` to `EnvKeyStrategy.ENV` if not set
    env_key_strat: EnvKeyStrategy | None = meta.load_case or EnvKeyStrategy.ENV
    default_strat = env_key_strat is not EnvKeyStrategy.STRICT
    # default `env_precedence` to SECRETS_ENV_DOTENV if not set
    env_precedence: EnvPrecedence = meta.env_precedence or EnvPrecedence.SECRETS_ENV_DOTENV

    field_to_env_vars = resolve_dataclass_field_to_env_for_load(cls)
    check_env_vars = True if field_to_env_vars else False

    field_to_paths = DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD[cls]
    has_alias_paths = True if field_to_paths else False

    # FIXME get from functions instead
    has_defaults = SEEN_DEFAULT[cls]

    # Fix for using `auto_assign_tags` and `on_unknown_key='RAISE'` together
    # See https://github.com/rnag/dataclass-wizard/issues/137
    has_tag_assigned = getattr(meta, 'tag', None) is not None
    if (has_tag_assigned and
            # Ensure `tag_key` isn't a dataclass field,
            # to avoid issues with our logic.
            # See https://github.com/rnag/dataclass-wizard/issues/148
            meta.tag_key not in cls_init_field_names):
        expect_tag_as_unknown_key = True
    else:
        expect_tag_as_unknown_key = False

    # on_unknown_key = meta.on_unknown_key

    catch_all_field: str | None = field_to_env_vars.pop(CATCH_ALL, None)
    has_catch_all = catch_all_field is not None

    if has_catch_all:
        pre_assign = 'i+=1; '
        catch_all_field_stripped = catch_all_field.rstrip('?')
        catch_all_idx = cls_init_field_names.index(catch_all_field_stripped)
        # remove catch all field from list, so we don't iterate over it
        del cls_init_fields[catch_all_idx]
    else:
        pre_assign = ''
        catch_all_field_stripped = catch_all_idx = None

    # if on_unknown_key is not None:
    #     should_raise = on_unknown_key is KeyAction.RAISE
    #     should_warn = on_unknown_key is KeyAction.WARN
    #     if should_warn or should_raise:
    #         pre_assign = 'i+=1; '
    #         set_aliases = True
    #     else:
    #         set_aliases = has_catch_all
    # else:
    #     should_raise = should_warn = None
    set_aliases = has_catch_all

    if set_aliases:
        if expect_tag_as_unknown_key:
            # add an alias for the tag key, so we don't
            # capture or raise an error when we see it
            aliases = {meta.tag_key}
        else:
            aliases = set()

        new_locals['aliases'] = aliases
    else:
        aliases = None

    if has_alias_paths:
        new_locals['safe_get'] = env_safe_get

    add_body_lines = cls_init_fields or has_catch_all

    _env_defaults: EnvInit = {}
    if _env_file := meta.env_file:
        _env_defaults['file'] = _env_file
    if (_env_prefix := meta.env_prefix) is not None:
        _env_defaults['prefix'] = _env_prefix
    LOG.debug('__env__ defaults = %r', _env_defaults)
    if (_secrets_dir := meta.secrets_dir) is not None:
        _env_defaults['secrets_dir'] = _secrets_dir
        LOG.debug('secrets_dir = <configured>')

    new_locals['_env_defaults'] = _env_defaults
    init_params = ['self',
                   "__env__:'EnvInit'=None",
                   '*']

    with fn_gen.function(fn_name, init_params, MISSING, new_locals):
        if add_body_lines:
            fn_gen.add_line('cfg = _env_defaults if __env__ is None else _env_defaults | __env__')
            fn_gen.add_line("reload = cfg.get('reload', False)")
            fn_gen.add_line("env_file = cfg.get('file')")
            fn_gen.add_line("secrets_dir = cfg.get('secrets_dir')")
            fn_gen.add_line("pfx = cfg.get('prefix', '')")
            # Need to create a separate dictionary to copy over the constructor
            # args, as we don't want to mutate the original dictionary object.
            if pre_assign:
                fn_gen.add_line('i = 0')

            env_map_assign = "cfg.get('mapping') or os.environ"
            if env_precedence is EnvPrecedence.ENV_ONLY:
                fn_gen.add_line(f'env = {env_map_assign}')
            else:
                fn_gen.add_line(f'env_map = {env_map_assign}')

                order = _PRECEDENCE_ORDER[env_precedence]

                fn_gen.add_line('maps = []')
                fn_gen.add_line(f'# precedence: {env_precedence.value}')
                for src in order:
                    if src == 'secrets':
                        with fn_gen.if_('secrets_dir is not None'):
                            fn_gen.add_line('maps.append(cls._dcw_env_cache_secrets(secrets_dir, reload=reload))')
                    elif src == 'dotenv':
                        with fn_gen.if_('env_file'):
                            fn_gen.add_line('maps.append(cls._dcw_env_cache_dotenv(env_file, reload=reload))')
                    elif src == 'env':
                        fn_gen.add_line('maps.append(env_map)')

                fn_gen.add_line('env = env_map if len(maps) == 1 else ChainMap(*maps)')

            if (_pre_from_dict := getattr(cls, '_pre_from_dict', None)) is not None:
                new_locals['__pre_from_dict__'] = _pre_from_dict
                fn_gen.add_line('env = __pre_from_dict__(env)')

            fn_gen.add_line('_vars = None')

            with fn_gen.try_():

                if expect_tag_as_unknown_key and pre_assign:
                    with fn_gen.if_(f'{meta.tag_key!r} in env'):
                        fn_gen.add_line('i+=1')

                val = 'v1'
                _val_is_found = f'{val} is not MISSING'
                for i, f in enumerate(cls_init_fields):
                    name = f.name
                    preferred_env_var = f"f'{{pfx}}{name}'"
                    has_default = has_defaults and (
                        f.default is not MISSING
                        or f.default_factory is not MISSING
                    )
                    val_is_found = _val_is_found

                    tp_var = f'tp_{i}'
                    new_locals[tp_var] = f.type

                    init_params.append(f'{name}:{tp_var}=MISSING')

                    f_assign = f'field={name!r}; {val}={name}'

                    condition = [val_is_found]

                    if (has_alias_paths
                            and (paths := field_to_paths.get(name)) is not None):

                        if len(paths) == 1:
                            first_key, *path = paths[0]

                            # add the first part (top-level key) of the path
                            if set_aliases:
                                aliases.add(first_key)

                            condition.append(
                                f'({val} := safe_get(env, {first_key!r}, {path!r}, {not has_default})) is not MISSING'
                            )
                        else:
                            last_idx = len(paths) - 1

                            for k, path in enumerate(paths):
                                first_key, *path = path

                                # add the first part (top-level key) of each path
                                if set_aliases:
                                    aliases.add(first_key)

                                if k == last_idx:
                                    condition.append(
                                        f'({val} := safe_get(env, {first_key!r}, {path!r}, {not has_default})) is not MISSING')
                                else:
                                    condition.append(
                                        f'({val} := safe_get(env, {first_key!r}, {path!r}, False)) is not MISSING')

                        # TODO raise some useful message like (ex. on IndexError):
                        #       Field "my_str" of type tuple[float, str] in A2 has invalid value ['123']

                    else:
                        if (check_env_vars
                            and (_initial_env_vars := field_to_env_vars.get(name)) is not None):
                            if len(_initial_env_vars) == 1:
                                _aliases = [_initial_env_vars[0]]
                            else:
                                _aliases = list(_initial_env_vars)
                            _has_alias = True
                            # No prefix for explicit aliases!
                            condition.extend([
                                f'({val} := env.get({alias!r}, MISSING)) is not MISSING'
                                for alias in _initial_env_vars
                            ])
                            preferred_env_var = repr(_initial_env_vars[0])
                        else:
                            _aliases = []
                            _has_alias = False

                        if default_strat:
                            _env_vars = possible_env_vars(name, env_key_strat)
                            condition.extend([
                                f"({val} := env.get(f'{{pfx}}{alias}', MISSING)) is not MISSING"
                                for alias in _env_vars
                            ])
                            _aliases.extend(_env_vars)
                            if not _has_alias:
                                preferred_env_var = f"f'{{pfx}}{_env_vars[0]}'"
                        else:  # EnvKeyStrategy.STRICT
                            pass

                        if set_aliases:
                            # add field name itself
                            aliases.add(name)
                            # add possible JSON keys
                            aliases.update(_aliases)

                    if len(condition) > 1:
                        val_is_found = '(' + '\n     or '.join(condition) + ')'
                    else:
                        val_is_found = condition[0]

                    string = generate_field_code(cls_loader, extras, f, i)

                    if f_assign is not None:
                        fn_gen.add_line(f_assign)

                    if has_default:
                        with fn_gen.if_(val_is_found):
                            fn_gen.add_line(f'{pre_assign}self.{name} = {string}')
                        if (default_factory := f.default_factory) is not MISSING:
                            with fn_gen.else_():
                                default_factory_name = f'_dflt{i}'
                                new_locals[default_factory_name] = default_factory
                                fn_gen.add_line(f'self.{name} = {default_factory_name}()')

                    else:
                        with fn_gen.if_(val_is_found):
                            fn_gen.add_line(f'{pre_assign}self.{name} = {string}')
                        with fn_gen.else_():
                            fn_gen.add_line(f'_vars = add(_vars, field, {preferred_env_var}, {tp_var})')

                # check for any required fields with missing values
                with fn_gen.if_('_vars is not None'):
                    fn_gen.add_line('raise MissingVars(cls, _vars) from None')

            # create a broad `except Exception` block, as we will be
            # re-raising all exception(s) as a custom `ParseError`.
            with fn_gen.except_(Exception, 'e', ParseError):
                fn_gen.add_line("re_raise(e, cls, env, fields, field, locals().get('v1'))")

        elif not has_post_init:
            fn_gen.add_line('pass')

        if not cls_init_fields:
            init_params.pop()  # remove trailing `*` in function params

        if has_catch_all:
            catch_all_def = '{k: env[k] for k in env if k not in aliases}'

            if catch_all_field.endswith('?'):  # Default value
                with fn_gen.if_('len(env) != i'):
                    fn_gen.add_line(f'self.{catch_all_field_stripped} = {catch_all_def}')
            else:
                fn_gen.add_line(f'self.{catch_all_field_stripped} = {{}} if len(env) == i else {catch_all_def}')

        # elif set_aliases:  # warn / raise on unknown key
        #     line = 'extra_keys = set(env) - aliases'
        #
        #     with fn_gen.if_('len(env) != i'):
        #         fn_gen.add_line(line)
        #         if should_raise:
        #             # Raise an error here (if needed)
        #             new_locals['UnknownKeysError'] = UnknownKeysError
        #             fn_gen.add_line('raise UnknownKeysError(extra_keys, env, cls, fields) from None')
        #         elif should_warn:
        #             # Show a warning here
        #             new_locals['LOG'] = LOG
        #             fn_gen.add_line(r"LOG.warning('Found %d unknown keys %r not mapped to the dataclass schema.\n"
        #                             r"  Class: %r\n  Dataclass fields: %r', "
        #                             "len(extra_keys), extra_keys, "
        #                             "cls.__qualname__, [f.name for f in fields])")

        # Does this class have a post-init function?
        if has_post_init:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            params_str = ','.join(f.name for f in fields
                                  if f._field_type is _FIELD_INITVAR)
            fn_gen.add_line(f'self.{_POST_INIT_NAME}({params_str})')

    # Save the load function for the main dataclass, so we don't need to run
    # this logic each time.
    _locals = {}
    with fn_gen.function(raw_dict_name, ['self'], JSONObject, _locals):
        parts = ','.join([f'{name!r}:self.{name}' for name in cls.__field_names__])
        fn_gen.add_line(f'return {{{parts}}}')

    # noinspection PyUnboundLocalVariable
    functions = fn_gen.create_functions(_globals)

    cls_init = functions[fn_name]
    cls_raw_dict = functions[raw_dict_name]

    set_new_attribute(
        cls, '__init__', cls_init)
    LOG.debug("setattr(%s, '__init__', %s)",
              cls_name, fn_name)

    set_new_attribute(
        cls, 'raw_dict', cls_raw_dict)
    LOG.debug("setattr(%s, 'raw_dict', %s)",
              cls_name, raw_dict_name)

    return cls_init


def _add_missing_var(missing_vars: dict | None, name, var_name, tp):
    tn = type_name(tp)

    # noinspection PyBroadException
    try:
        suggested = tp()
    except Exception:
        suggested = None

    if missing_vars is None:
        missing_vars = []

    missing_vars.append((name, var_name, tn, suggested))

    return missing_vars


# def _handle_parse_error(e, cls, name, env_prefix, var_name):
#
#     # We run into a parsing error while loading the field
#     # value; Add additional info on the Exception object
#     # before re-raising it.
#     e.class_name = cls
#     e.field_name = name
#     e.kwargs['env_variable'] = _get_var_name(name, env_prefix, var_name)
#
#     raise


def generate_field_code(cls_loader: LoadMixin,
                        extras: Extras,
                        field: Field,
                        field_i: int) -> str | TypeInfo:

    cls = extras['cls']
    field_type = field.type = eval_forward_ref_if_needed(field.type, cls)

    try:
        return cls_loader.load_dispatcher_for_annotation(
            TypeInfo(field_type, field_i=field_i), extras
        )

    # except Exception as e:
    #     re_raise(e, cls, None, dataclass_init_fields(cls), field, None)
    except ParseError as pe:
        pe.class_name = cls
        # noinspection PyPropertyAccess
        pe.field_name = field.name
        raise pe from None


def re_raise(e, cls, o, fields, field, value):
    # If the object `o` is None, then raise an error with
    # the relevant info included.
    if o is None:
        raise MissingData(cls) from None

    # Check if the object `o` is some other type than what we expect -
    # for example, we could be passed in a `list` type instead.
    if not isinstance(o, Mapping):
        base_err = TypeError('Incorrect type for `from_dict()`')
        e = ParseError(base_err, o, dict, cls, desired_type=dict)

    add_fields = True
    if type(e) is not ParseError:
        if isinstance(e, JSONWizardError):
            add_fields = False
        else:
            tp = getattr(next((f for f in fields if f.name == field), None), 'type', Any)
            e = ParseError(e, value, tp, 'load')

    # We run into a parsing error while loading the field value;
    # Add additional info on the Exception object before re-raising it.
    #
    # First confirm these values are not already set by an
    # inner dataclass. If so, it likely makes it easier to
    # debug the cause. Note that this should already be
    # handled by the `setter` methods.
    if add_fields:
        e.class_name, e.fields, e.field_name, e.json_object = cls, fields, field, o
    else:
        e.class_name, e.field_name, e.json_object = cls, field, o

    raise e from None


class LoadMixin(V1LoadMixin, _setup_defaults=False):
    """
    This Mixin class derives its name from the eponymous `json.loads`
    function. Essentially it contains helper methods to convert JSON strings
    (or a Python dictionary object) to a `dataclass` which can often contain
    complex types such as lists, dicts, or even other dataclasses nested
    within it.

    Refer to the :class:`AbstractLoader` class for documentation on any of the
    implemented methods.

    """
    __slots__ = ()

    @staticmethod
    def is_none(tp: TypeInfo, extras: Extras) -> str:
        o = tp.v()
        return f"{o} is None or {o} == 'null'"

    @staticmethod
    def load_to_bytes(tp: TypeInfo, extras: Extras):
        # could add support for b64-encoded strings later:
        # bytes(__b64decode(o)) if (o.__class__ is str and __env_b64)
        o = tp.v()
        return (f"{o} if (t := {o}.__class__) is bytes "
                f"else {o}.encode('utf-8') if t is str "
                f"else bytes({o})")

    @classmethod
    def load_to_bytearray(cls, tp: TypeInfo, extras: Extras):
        o = tp.v()
        as_bytes = cls.load_to_bytes(tp, extras)
        return (f'{o} if {o}.__class__ is bytearray '
                f'else {tp.wrap_builtin(bytearray, as_bytes, extras)}')

    @classmethod
    def load_to_dataclass(cls, tp: TypeInfo, extras: Extras):
        # pre-decoder wraps `v()` in `asdict(...)`, so use the wrapped value
        o = tp.v_for_def()
        tn = tp.type_name(extras)
        from_dict = super().load_to_dataclass(tp, extras)
        return f'{o} if {o}.__class__ is {tn} else {from_dict}'



import logging
from dataclasses import dataclass, field, asdict
from datetime import date, datetime
from timeit import timeit
from typing import TypeVar, List, Union

import dataclass_factory
import marshmallow
import pytest
from dataclasses_json import DataClassJsonMixin, config
from jsons import JsonSerializable
from dacite import from_dict as dacite_from_dict
from pydantic import BaseModel
import mashumaro

from dataclass_wizard import JSONWizard, LoadMeta
from dataclass_wizard._type_utils import create_new_class
from dataclass_wizard.constants import PY314_OR_ABOVE
from dataclass_wizard.utils._string_case import to_snake_case
# FIXME
from dataclass_wizard.cli.schema import (
    _as_datetime as as_datetime,
    _as_date as as_date)


log = logging.getLogger(__name__)

# Dataclass Definitions (Same as before, no changes needed)
@dataclass
class Data1:
    instance: 'Instance'
    result: 'Result'


@dataclass
class Instance:
    name: str
    data: 'Data2'


@dataclass
class Data2:
    date: date
    owner: str


@dataclass
class Result:
    status: str
    iteration_results: 'IterationResults'


@dataclass
class IterationResults:
    iterations: List['Iteration']


@dataclass
class Iteration:
    name: str
    data: 'Data3'


@dataclass
class Data3:
    question1: str
    question2: str


# New Model Class Definitions for Libraries

class MyClassPydantic(BaseModel):
    instance: 'InstancePydantic'
    result: 'ResultPydantic'


class InstancePydantic(BaseModel):
    name: str
    data: 'Data2Pydantic'


class Data2Pydantic(BaseModel):
    date: date
    owner: str


class ResultPydantic(BaseModel):
    status: str
    iteration_results: 'IterationResultsPydantic'


@dataclass
class IterationResultsPydantic:
    iterations: List['IterationPydantic']


class IterationPydantic(BaseModel):
    name: str
    data: 'Data3Pydantic'


class Data3Pydantic(BaseModel):
    question1: str
    question2: str


@dataclass
class MyClassMashumaro(mashumaro.DataClassDictMixin):
    instance: 'InstanceMashumaro'
    result: 'Result'


@dataclass
class InstanceMashumaro:
    name: str
    data: 'Data2Mashumaro'


@dataclass
class Data2Mashumaro:
    date: date
    owner: str


# Corrected Definition for `MyClassDJ`
@dataclass
class MyClassDJ(DataClassJsonMixin):
    instance: 'InstanceDJ'
    result: 'Result'


class InstanceDJ:
    name: str
    data: 'Data2DJ'


class Data2DJ:
    date: date
    owner: str


# Model for `dataclass-wizard`
WizType = TypeVar('WizType', Data1, JSONWizard)
# Model for `jsons`
JsonsType = TypeVar('JsonsType', Data1, JsonSerializable)
# Model for `dataclasses-json`
DJType = TypeVar('DJType', Data1, DataClassJsonMixin)
# Model for `mashumaro`
MashumaroType = TypeVar('MashumaroType', Data1, mashumaro.DataClassDictMixin)
# Factory for `dataclass-factory`
factory = dataclass_factory.Factory()

MyClassWizard: WizType = create_new_class(
    Data1, (Data1, JSONWizard), 'Wizard',
    attr_dict=vars(Data1).copy())
MyClassJsons: JsonsType = create_new_class(
    Data1, (Data1, JsonSerializable), 'Jsons',
    attr_dict=vars(Data1).copy())
MyClassMashumaroModel: MashumaroType = create_new_class(
    Data1, (Data1, mashumaro.DataClassDictMixin), 'Mashumaro',
    attr_dict=vars(Data1).copy())

# Pydantic Model for Benchmarking
MyClassPydanticModel = MyClassPydantic

# Mashumaro Model for Benchmarking
# MyClassMashumaroModel = MyClassMashumaro


# Enable experimental `v1` mode for optimized de/serialization
LoadMeta(v1=True).bind_to(MyClassWizard)


@pytest.fixture(scope='session')
def data():
    return {
        "instance": {
            "name": "example1",
            "data": {
                "date": "2021-01-01",
                "owner": "Maciek"
            }
        },
        "result": {
            "status": "complete",
            "iteration_results": {
                "iterations": [
                    {
                        "name": "first",
                        "data": {
                            "question1": "yes",
                            "question2": "no"
                        }
                    }
                ]
            }
        }
    }


dt_iso_format_schema = dataclass_factory.Schema(
    parser=as_datetime,
    serializer=datetime.isoformat
)

date_iso_format_schema = dataclass_factory.Schema(
    parser=as_date,
    serializer=date.isoformat
)

factory.schemas = {
    datetime: dt_iso_format_schema,
    date: date_iso_format_schema
}


def test_load(request, data, n):
    """
    [ RESULTS]
    platform darwin -- Python 3.13.11, pytest-8.3.4, pluggy-1.6.0

    benchmarks.nested.nested - [INFO] dataclass-wizard     0.128877
    benchmarks.nested.nested - [INFO] dataclass-factory    0.405885
    benchmarks.nested.nested - [INFO] dataclasses-json     11.878780
    benchmarks.nested.nested - [INFO] mashumaro            0.154879
    benchmarks.nested.nested - [INFO] pydantic             0.286836
    benchmarks.nested.nested - [INFO] jsons                24.753070
    benchmarks.nested.nested - [INFO] jsons (strict)       26.192690

    """
    g = globals().copy()
    g.update(locals())

    MyClassWizard.from_dict(data)

    log.info('dataclass-wizard     %f',
             timeit('MyClassWizard.from_dict(data)', globals=g, number=n))

    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        log.info('dataclass-factory    %f',
                 timeit('factory.load(data, Data1)', globals=g, number=n))

    log.info('dataclasses-json     %f',
             timeit('MyClassDJ.from_dict(data)', globals=g, number=n))

    # JUST SKKIPING IN INTERESTS OF TIME
    # log.info('dacite               %f',
    #          timeit('dacite_from_dict(MyClass, data)', globals=g, number=n))

    log.info('mashumaro            %f',
             timeit('MyClassMashumaro.from_dict(data)', globals=g, number=n))

    log.info('pydantic             %f',
             timeit('MyClassPydantic(**data)', globals=g, number=n))

    if not request.config.getoption("--all"):
        pytest.skip("Skipping benchmarks for the rest by default, unless --all is specified.")

    log.info('jsons                %f',
             timeit('MyClassJsons.load(data)', globals=g, number=n))

    log.info('jsons (strict)       %f',
             timeit('MyClassJsons.load(data, strict=True)', globals=g, number=n))

    c1 = MyClassWizard.from_dict(data)
    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        c2 = factory.load(data, Data1)
    c3 = MyClassDJ.from_dict(data)
    c4 = MyClassJsons.load(data)
    c5 = MyClassMashumaro.from_dict(data)
    # c6 = dacite_from_dict(MyClass, data)
    c7 = MyClassPydantic(**data)

    assert c1.__dict__ == c3.__dict__ == c4.__dict__ == c5.__dict__ == c7.__dict__ # == c6.__dict__

    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        assert c1.__dict__ == c2.__dict__


def test_dump(request, data, n):
    """
    [ RESULTS]
    platform darwin -- Python 3.13.11, pytest-8.3.4, pluggy-1.6.0

    benchmarks.nested.nested - [INFO] dataclass-wizard     0.093090
    benchmarks.nested.nested - [INFO] asdict (dataclasses) 0.621261
    benchmarks.nested.nested - [INFO] dataclass-factory    0.209332
    benchmarks.nested.nested - [INFO] dataclasses-json     5.172026
    benchmarks.nested.nested - [INFO] mashumaro            0.074544
    benchmarks.nested.nested - [INFO] pydantic             0.174419
    benchmarks.nested.nested - [INFO] jsons                40.467886
    benchmarks.nested.nested - [INFO] jsons (strict)       36.541698
    """
    c1 = MyClassWizard.from_dict(data)
    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        c2 = factory.load(data, Data1)
    c3 = MyClassDJ.from_dict(data)
    c4 = MyClassJsons.load(data)
    c5 = MyClassMashumaro.from_dict(data)
    c6 = MyClassPydantic(**data)

    g = globals().copy()
    g.update(locals())

    log.info('dataclass-wizard     %f',
             timeit('c1.to_dict()', globals=g, number=n))

    log.info('asdict (dataclasses) %f',
             timeit('asdict(c1)', globals=g, number=n))

    if not PY314_OR_ABOVE:  # breaks on Python 3.14+
        log.info('dataclass-factory    %f',
                 timeit('factory.dump(c2, Data1)', globals=g, number=n))

    log.info('dataclasses-json     %f',
             timeit('c3.to_dict()', globals=g, number=n))

    log.info('mashumaro            %f',
             timeit('c5.to_dict()', globals=g, number=n))

    log.info('pydantic             %f',
             timeit('c6.model_dump()', globals=g, number=n))

    if not request.config.getoption("--all"):
        pytest.skip("Skipping benchmarks for the rest by default, unless --all is specified.")

    log.info('jsons                %f',
             timeit('c4.dump()', globals=g, number=n))

    log.info('jsons (strict)       %f',
             timeit('c4.dump(strict=True)', globals=g, number=n))

    # Assert the dict objects which are the result of `to_dict` are all equal.
    c1_dict = {to_snake_case(f): fval for f, fval in c1.to_dict().items()}

    # assert c1_dict == factory.dump(c2, Data1) == c3.to_dict() == c4.dump() == c5.to_dict()



from typing import AnyStr, BinaryIO

from .._serial_json import SerializerHookMixin
from .._type_def import (
    Decoder,
    Encoder,
    FileDecoder,
    FileEncoder,
    FileType,
    ParseFloat,
    T,
)

class TOMLWizard(SerializerHookMixin):

    def __init_subclass__(cls, dump_case=None):
        ...

    @classmethod
    def from_toml(cls: type[T],
                  string_or_stream: AnyStr | BinaryIO, *,
                  decoder: Decoder | None = None,
                  header: str = 'items',
                  parse_float: ParseFloat = float) -> T | list[T]:
        ...

    @classmethod
    def from_toml_file(cls: type[T], file: FileType, *,
                       decoder: FileDecoder | None = None,
                       header: str = 'items',
                       parse_float: ParseFloat = float) -> T | list[T]:
        ...

    def to_toml(self: T,
                /,
                *encoder_args,
                encoder: Encoder | None = None,
                multiline_strings: bool = False,
                indent: int = 4) -> str:
        ...

    def to_toml_file(self: T, file: FileType, mode: str = 'wb',
                     encoder: FileEncoder | None = None,
                     multiline_strings: bool = False,
                     indent: int = 4) -> None:
        ...

    @classmethod
    def list_to_toml(cls: type[T],
                     instances: list[T],
                     header: str = 'items',
                     encoder: Encoder | None = None,
                     **encoder_kwargs) -> str:
        ...



from dataclasses import MISSING
from typing import Any, Sequence, TypeAlias, Union

PathPart: TypeAlias = Union[str, int, float, bool]
PathType: TypeAlias = Sequence[PathPart]


def safe_get(data: dict | list,
             path: PathType,
             default=MISSING,
             raise_: bool = True) -> Any:
    """
    Retrieve a value from a nested structure safely.

    Traverses a nested structure (e.g., dictionaries or lists) following a sequence of keys or indices specified in `path`.
    Handles missing keys, out-of-bounds indices, or invalid types gracefully.

    Args:
        data (Any): The nested structure to traverse.
        path (Iterable): A sequence of keys or indices to follow.
        default (Any): The value to return if the path cannot be fully traversed.
                       If not provided and an error occurs, the exception is re-raised.
        raise_ (bool): True to raise an error on invalid path (default True).

    Returns:
        Any: The value at the specified path, or `default` if traversal fails.

    Raises:
        KeyError, IndexError, AttributeError, TypeError: If `default` is not provided
        and an error occurs during traversal.
    """
    ...


def v1_safe_get(data: dict | list,
                path: PathType,
                raise_: bool) -> Any:
    """
    Retrieve a value from a nested structure safely.

    Traverses a nested structure (e.g., dictionaries or lists) following a sequence of keys or indices specified in `path`.
    Handles missing keys, out-of-bounds indices, or invalid types gracefully.

    Args:
        data (Any): The nested structure to traverse.
        path (Iterable): A sequence of keys or indices to follow.
        raise_ (bool): True to raise an error on invalid path.

    Returns:
        Any: The value at the specified path, or `MISSING` if traversal fails.

    Raises:
        KeyError, IndexError, AttributeError, TypeError: If `default` is not provided
        and an error occurs during traversal.
    """
    ...


def v1_env_safe_get(data: dict | list,
                    first_key: PathPart,
                    path: PathType,
                    raise_: bool) -> Any:
    """
    Retrieve a value from a nested structure safely.

    Traverses a nested structure (e.g., dictionaries or lists) following a sequence of keys or indices specified in `path`.
    Handles missing keys, out-of-bounds indices, or invalid types gracefully.

    Args:
        data (Any): The nested structure to traverse.
        path (Iterable): A sequence of keys or indices to follow.
        raise_ (bool): True to raise an error on invalid path.

    Returns:
        Any: The value at the specified path, or `MISSING` if traversal fails.

    Raises:
        KeyError, IndexError, AttributeError, TypeError: If `default` is not provided
        and an error occurs during traversal.
    """
    ...


def _format_err(e: Exception,
                current_data: Any,
                path: PathType,
                current_path: PathPart):
    """Format and return a `ParseError`."""
    ...


def split_object_path(_input: str) -> PathType:
    """
    Parse a custom object path string into a list of components.

    This function interprets a custom object path syntax and breaks it into individual path components,
    including dictionary keys, list indices, attributes, and nested elements.
    It handles escaped characters and supports mixed types (e.g., strings, integers, floats, booleans).

    Args:
        _input (str): The object path string to parse.

    Returns:
        PathType: A list of components representing the parsed path. Components can be strings,
                  integers, floats, booleans, or other valid key/index types.

    Example:
        >>> split_object_path(r'''a[b][c]["d\\\"o\\\""][e].f[go]['1'].then."y\\e\\\"s"[1]["we can!"].five.2.3.[ok][4.56].[-7.89].'let\\'sd\\othisy\\'all!'.yeah.123.False['True'].thanks!''')
        ['a', 'b', 'c', 'd"o"', 'e', 'f', 'go', '1', 'then', 'y\\e"s', 1, 'we can!', 'five', 2, 3, 'ok', 4.56, -7.89,
         "let'sd\\othisy'all!", 'yeah', 123, False, 'True', 'thanks!']
    """



__all__ = [
    'PY310_OR_ABOVE',
    'PY311_OR_ABOVE',
    'PY312_OR_ABOVE',
    'TypedDict',
    # For compatibility with Python 3.9 and 3.10
    'Required',
    'NotRequired',
    'ReadOnly',
    'LiteralString',
]

import sys

from dataclass_wizard.constants import PY313_OR_ABOVE

# Check if we are running Python 3.10+
PY310_OR_ABOVE = sys.version_info[:2] >= (3, 10)

# Check if we are running Python 3.11+
PY311_OR_ABOVE = sys.version_info[:2] >= (3, 11)

# Check if we are running Python 3.12+
PY312_OR_ABOVE = sys.version_info[:2] >= (3, 12)

# Check if we are running Python 3.9 or 3.10
PY310_OR_EARLIER = not PY311_OR_ABOVE

# Weird, test cases for `TypedDict` fail in Python 3.9 & 3.10.15 (3.10:latest)
# So apparently, we need to use the one from `typing_extensions`.
if PY310_OR_EARLIER:
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

# typing.Required and typing.NotRequired: Introduced in Python 3.11
if PY311_OR_ABOVE:
    from typing import Required
    from typing import NotRequired
    from typing import LiteralString
else:
    from typing_extensions import Required
    from typing_extensions import NotRequired
    from typing_extensions import LiteralString

if PY313_OR_ABOVE:
    from typing import ReadOnly
else:
    from typing_extensions import ReadOnly



import pytest
from pytest_mock import MockerFixture

from dataclass_wizard.v0 import fromlist
from dataclass_wizard.v0.models import Container, json_field
from .conftest import SampleClass


@pytest.fixture
def mock_open(mocker: MockerFixture):
    return mocker.patch('dataclass_wizard.v0.models.open')


def test_json_field_does_not_allow_both_default_and_default_factory():
    """
    Confirm we can't specify both `default` and `default_factory` when
    calling the :func:`json_field` helper function.
    """
    with pytest.raises(ValueError):
        _ = json_field((), default=None, default_factory=None)


def test_container_with_incorrect_usage():
    """Confirm an error is raised when wrongly instantiating a Container."""
    c = Container()

    with pytest.raises(TypeError) as exc_info:
        _ = c.to_json()

    err_msg = exc_info.exconly()
    assert 'A Container object needs to be instantiated ' \
           'with a generic type T' in err_msg


def test_container_methods(mocker: MockerFixture, mock_open):
    list_of_dict = [{'f1': 'hello', 'f2': 1},
                    {'f1': 'world', 'f2': 2}]

    list_of_a = fromlist(SampleClass, list_of_dict)

    c = Container[SampleClass](list_of_a)

    # The repr() is very short, so it would be expected to fit in one line,
    # which thus aligns with the output of `pprint.pformat`.
    assert str(c) == repr(c)

    assert c.prettify() == """\
[
  {
    "f1": "hello",
    "f2": 1
  },
  {
    "f1": "world",
    "f2": 2
  }
]"""

    assert c.to_json() == '[{"f1": "hello", "f2": 1}, {"f1": "world", "f2": 2}]'

    mock_open.assert_not_called()
    mock_encoder = mocker.Mock()

    filename = 'my_file.json'
    c.to_json_file(filename, encoder=mock_encoder)

    mock_open.assert_called_once_with(filename, 'w')
    mock_encoder.assert_called_once_with(list_of_dict, mocker.ANY)



from __future__ import annotations

from dataclasses import MISSING
from weakref import WeakKeyDictionary, WeakSet

from ._type_def import ExplicitNull
from ._type_utils import get_class, get_class_name, per_cls
from .constants import CATCH_ALL, PACKAGE_NAME
from .errors import InvalidConditionError
from .models import CatchAll, Field
from .utils._dataclass_compat import SEEN_DEFAULT, dataclass_fields
from .utils._typing_compat import (
    eval_forward_ref_if_needed,
    get_args,
    is_annotated,
)

# A mapping of dataclass to its loader.
CLASS_TO_LOADER = WeakKeyDictionary()

# A mapping of dataclass to its dumper.
CLASS_TO_DUMPER = WeakKeyDictionary()

# We use a sentinel mapping to confirm if we need to set up the load
# config for a dataclass on an initial run.
IS_CONFIG_SETUP = WeakSet()

# Load: A cached mapping, per dataclass, of instance field name to alias path
DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD = WeakKeyDictionary()

# Dump: A cached mapping, per dataclass, of instance field name to alias path
DATACLASS_FIELD_TO_ALIAS_PATH_FOR_DUMP = WeakKeyDictionary()

# Load: A cached mapping, per dataclass, of instance field name to alias
DATACLASS_FIELD_TO_ALIAS_FOR_LOAD = WeakKeyDictionary()

# Load: A cached mapping, per dataclass, of instance field name to env var
DATACLASS_FIELD_TO_ENV_FOR_LOAD = WeakKeyDictionary()

# Dump: A cached mapping, per dataclass, of instance field name to alias
DATACLASS_FIELD_TO_ALIAS_FOR_DUMP = WeakKeyDictionary()

# A cached mapping, per dataclass, of instance field name to `SkipIf`
# condition
DATACLASS_FIELD_TO_SKIP_IF = WeakKeyDictionary()

# Cache: owner class -> its `Meta` inner class (only present when subclassed)
META_INITIALIZER = {}


def set_class_loader(cls_to_loader, class_or_instance, loader):

    cls = get_class(class_or_instance)
    loader_cls = get_class(loader)

    cls_to_loader[cls] = loader_cls

    return loader_cls


def set_class_dumper(cls_to_dumper, class_or_instance, dumper):

    cls = get_class(class_or_instance)
    dumper_cls = get_class(dumper)

    cls_to_dumper[cls] = dumper_cls

    return dumper_cls


def dataclass_field_to_skip_if(cls):
    return per_cls(DATACLASS_FIELD_TO_SKIP_IF, cls)


def resolve_dataclass_field_to_alias_for_dump(cls):

    if cls not in IS_CONFIG_SETUP:
        setup_config_for_cls(cls)

    return DATACLASS_FIELD_TO_ALIAS_FOR_DUMP[cls]


def resolve_dataclass_field_to_alias_for_load(cls):

    if cls not in IS_CONFIG_SETUP:
        setup_config_for_cls(cls)

    return DATACLASS_FIELD_TO_ALIAS_FOR_LOAD[cls]


def resolve_dataclass_field_to_env_for_load(cls):

    if cls not in IS_CONFIG_SETUP:
        setup_config_for_cls(cls)

    return DATACLASS_FIELD_TO_ENV_FOR_LOAD[cls]


def _process_field(name: str,
                   f: Field,
                   set_paths: bool,
                   init: bool,
                   load_dataclass_field_to_path,
                   dump_dataclass_field_to_path,
                   load_dataclass_field_to_alias,
                   load_dataclass_field_to_env,
                   dump_dataclass_field_to_alias):
    """Process a :class:`Field` for a dataclass field."""

    if f.path is not None:
        if set_paths:
            if init and f.load_alias is not ExplicitNull:
                load_dataclass_field_to_path[name] = f.path
            if not f.skip and f.dump_alias is not ExplicitNull:
                dump_dataclass_field_to_path[name] = f.path[0]
        # TODO I forget why this is needed :o
        if f.skip:
            dump_dataclass_field_to_alias[name] = ExplicitNull
        elif f.dump_alias is not ExplicitNull:
            dump_dataclass_field_to_alias[name] = ''

    else:
        if init:
            if f.load_alias is not None:
                load_dataclass_field_to_alias[name] = f.load_alias
            if f.env_vars is not None:
                load_dataclass_field_to_env[name] = f.env_vars
        if f.skip:
            dump_dataclass_field_to_alias[name] = ExplicitNull
        elif (dump := f.dump_alias) is not None:
            dump_dataclass_field_to_alias[name] = dump if isinstance(
                dump, str) else dump[0]



# Set up load and dump config for dataclass
def setup_config_for_cls(cls):
    load_dataclass_field_to_alias = per_cls(
        DATACLASS_FIELD_TO_ALIAS_FOR_LOAD, cls)
    load_dataclass_field_to_env = per_cls(
        DATACLASS_FIELD_TO_ENV_FOR_LOAD, cls)
    dump_dataclass_field_to_alias = per_cls(
        DATACLASS_FIELD_TO_ALIAS_FOR_DUMP, cls)

    dataclass_field_to_path = per_cls(
        DATACLASS_FIELD_TO_ALIAS_PATH_FOR_LOAD, cls)
    dump_dataclass_field_to_path = per_cls(
        DATACLASS_FIELD_TO_ALIAS_PATH_FOR_DUMP, cls)

    set_paths = False if dataclass_field_to_path else True
    field_to_skip_if = per_cls(DATACLASS_FIELD_TO_SKIP_IF, cls)
    seen_default = False

    for f in dataclass_fields(cls):
        init = f.init
        field_type = f.type = eval_forward_ref_if_needed(f.type, cls)

        if (init and not seen_default
            and (f.default is not MISSING
                 or f.default_factory is not MISSING)):
            seen_default = True

        # isinstance(f, Field) == True

        # Check if the field is a known `Field` subclass. If so, update
        # the class-specific mapping of JSON key to dataclass field name.
        if isinstance(f, Field):
            _process_field(f.name, f, set_paths, init,
                           dataclass_field_to_path,
                           dump_dataclass_field_to_path,
                           load_dataclass_field_to_alias,
                           load_dataclass_field_to_env,
                           dump_dataclass_field_to_alias)

        elif f.metadata:
            if value := f.metadata.get('__remapping__'):
                if isinstance(value, Field):
                    _process_field(f.name, value, set_paths, init,
                                   dataclass_field_to_path,
                                   dump_dataclass_field_to_path,
                                   load_dataclass_field_to_alias,
                                   load_dataclass_field_to_env,
                                   dump_dataclass_field_to_alias)
            elif value := f.metadata.get('__skip_if__'):
                if getattr(value, '__dcw_condition__', False):
                    field_to_skip_if[f.name] = value

        # Check for a "Catch All" field
        if field_type is CatchAll:
            load_dataclass_field_to_alias[CATCH_ALL] \
                = load_dataclass_field_to_env[CATCH_ALL] \
                = dump_dataclass_field_to_alias[CATCH_ALL] \
                = f'{f.name}{"" if f.default is MISSING else "?"}'

        # Check if the field annotation is an `Annotated` type. If so,
        # look for any `Field` objects in the arguments; for each object,
        # call `_process_field`.
        elif is_annotated(field_type):
            for extra in get_args(field_type)[1:]:
                if isinstance(extra, Field):
                    _process_field(f.name, extra, set_paths, init,
                                   dataclass_field_to_path,
                                   dump_dataclass_field_to_path,
                                   load_dataclass_field_to_alias,
                                   load_dataclass_field_to_env,
                                   dump_dataclass_field_to_alias)
                elif getattr(extra, '__dcw_condition__', False):
                    field_to_skip_if[f.name] = extra
                    if not getattr(extra, '_wrapped', False):
                        raise InvalidConditionError(cls, f.name) from None

    SEEN_DEFAULT[cls] = seen_default

    IS_CONFIG_SETUP.add(cls)


def call_meta_initializer_if_needed(cls, package_name=PACKAGE_NAME):
    """
    Calls the Meta initializer when the inner :class:`Meta` is sub-classed.
    """
    # TODO add tests

    # skip classes provided by this library
    if cls.__module__.startswith(f'{package_name}.'):
        return

    cls_name = get_class_name(cls)

    if cls_name in META_INITIALIZER:
        META_INITIALIZER[cls_name](cls)

    # Get the last immediate superclass
    base = cls.__base__

    # skip base `object` and classes provided by this library
    if (base is not object
            and not base.__module__.startswith(f'{package_name}.')):

        base_cls_name = get_class_name(base)

        if base_cls_name in META_INITIALIZER:
            META_INITIALIZER[base_cls_name](cls)



import os
from collections import namedtuple
from dataclasses import dataclass
from datetime import datetime, date, timezone
from typing import Tuple, NamedTuple, List

import pytest

from dataclass_wizard.v0 import EnvWizard
from dataclass_wizard.v0.environ.loaders import EnvLoader


def test_load_to_bytes():
    assert EnvLoader.load_to_bytes('testing 123', bytes) == b'testing 123'


@pytest.mark.parametrize(
    'input,expected',
    [
        ('testing 123', bytearray(b'testing 123')),
        (b'test', bytearray(b'test')),
        ([1, 2, 3], bytearray([1, 2, 3]))
    ]
)
def test_load_to_bytearray(input, expected):
    assert EnvLoader.load_to_byte_array(input, bytearray) == expected


def test_load_to_tuple_and_named_tuple():
    os.environ['MY_TUP'] = '1,2,3'
    os.environ['MY_NT'] = '[1.23, "string"]'
    os.environ['my_untyped_nt'] = 'hello ,  world, 123'

    class MyNT(NamedTuple):
        my_float: float
        my_str: str

    untyped_tup = namedtuple('untyped_tup', ('a', 'b', 'c'))

    class MyClass(EnvWizard, reload_env=True):
        my_tup: Tuple[int, ...]
        my_nt: MyNT
        my_untyped_nt: untyped_tup

    c = MyClass()

    assert c.dict() == {'my_nt': MyNT(my_float=1.23, my_str='string'),
                        'my_tup': (1, 2, 3),
                        'my_untyped_nt': untyped_tup(a='hello', b='world', c='123')}

    assert c.to_dict() == {'my_nt': MyNT(my_float=1.23, my_str='string'),
                           'my_tup': (1, 2, 3),
                           'my_untyped_nt': untyped_tup(a='hello', b='world', c='123')}


def test_load_to_dataclass():
    """When an `EnvWizard` subclass has a nested dataclass schema."""

    os.environ['inner_cls_1'] = 'my_bool=false, my_string=test'
    os.environ['inner_cls_2'] = '{"answerToLife": "42", "MyList": "testing, 123 , hello!"}'

    @dataclass
    class Inner1:
        my_bool: bool
        my_string: str

    @dataclass
    class Inner2:
        answer_to_life: int
        my_list: List[str]

    class MyClass(EnvWizard, reload_env=True):

        inner_cls_1: Inner1
        inner_cls_2: Inner2

    c = MyClass()
    # print(c)

    assert c.dict() == {
        'inner_cls_1': Inner1(my_bool=False,
                              my_string='test'),
        'inner_cls_2': Inner2(answer_to_life=42,
                              my_list=['testing', '123', 'hello!']),
    }

    assert c.to_dict() == {
        'inner_cls_1': {'my_bool': False,
                        'my_string': 'test'},
        'inner_cls_2': {'answer_to_life': 42,
                        'my_list': ['testing', '123', 'hello!']}
    }


@pytest.mark.parametrize(
    'input,expected',
    [
        ('2021-11-28T17:35:55', datetime(2021, 11, 28, 17, 35, 55)),
        (1577952245, datetime(2020, 1, 2, 8, 4, 5, tzinfo=timezone.utc)),
        (datetime.min, datetime.min)
    ]
)
def test_load_to_datetime(input, expected):
    assert EnvLoader.load_to_datetime(input, datetime) == expected


@pytest.mark.parametrize(
    'input,expected',
    [
        ('2021-11-28', date(2021, 11, 28)),
        (1577952245, date(2020, 1, 2)),
        (date.min, date.min)
    ]
)
def test_load_to_date(input, expected):
    assert EnvLoader.load_to_date(input, date) == expected



from logging import DEBUG, StreamHandler

from dataclass_wizard import DataclassWizard
from dataclass_wizard._meta_cache import get_meta


def test_dataclass_wizard_with_debug(restore_logger, mock_debug_log):
    """Subclass `DataclassWizard` with `debug=True`."""
    logger = restore_logger

    class _(DataclassWizard, debug=True):
        ...

    assert get_meta(_).debug == DEBUG

    assert logger.level == DEBUG
    assert logger.propagate is False
    assert any(isinstance(h, StreamHandler) for h in logger.handlers)
    # optional: ensure it didn't add duplicates
    assert sum(isinstance(h, StreamHandler) for h in logger.handlers) == 1



from os import PathLike, altsep, fspath, getcwd, sep
from os.path import isabs
from pathlib import Path

from ._lazy_imports import dotenv


def get_secrets_map(cls, secret_dirs, *, reload=False):
    if secret_dirs is None:
        return {}

    # Treat empty containers as empty; but empty string is likely a bug.
    if isinstance(secret_dirs, (str, PathLike)):
        if str(secret_dirs) == '':
            raise ValueError('secrets_dir cannot be an empty string')
        dirs = fspath(secret_dirs),
    else:
        dirs = tuple(fspath(d) for d in secret_dirs)

    # optionally normalize paths (realpath)
    key = dirs

    cache = getattr(cls, '_dw_secrets_cache', None)
    if cache is None:
        cache = cls._dw_secrets_cache = {}

    if not reload and key in cache:
        return cache[key]

    m = cache[key] = read_secrets_dirs(dirs)
    return m


def get_dotenv_map(cls, env_file, *, reload=False):
    if not env_file:
        return {}

    if env_file is True:
        files = '.env',
    elif isinstance(env_file, (str, PathLike)):
        files = fspath(env_file),
    else:
        files = tuple(fspath(f) for f in env_file)

    cwd = getcwd()
    key = (cwd, *files)

    cache = getattr(cls, '_dw_dotenv_cache', None)
    if cache is None:
        cache = cls._dw_dotenv_cache = {}

    if not reload and key in cache:
        return cache[key]

    m = cache[key] = dotenv_values(files)
    return m


def read_secrets_dirs(secret_dirs):
    out = {}

    for d in secret_dirs:
        if not isinstance(d, (str, PathLike)):
            raise TypeError(f'secrets_dir entries must be str/PathLike, got {type(d)!r}')

        p = Path(d)

        # Missing mount is common in Docker; treat as empty.
        if not p.exists():
            continue

        if p.is_file():
            raise ValueError(f'Secrets directory {p!r} is a file, not a directory.')
        if not p.is_dir():
            # broken symlink, device node, etc. -> ignore or raise; ignore is ok
            continue

        try:
            for f in p.iterdir():
                f: Path
                if not f.is_file():
                    continue
                # Docker secret files are typically single-line with trailing NL
                out[f.name] = f.read_text(encoding='utf-8').rstrip('\n')

        except OSError as e:
            # Permission issues, transient IO errors; choose raise vs ignore
            raise OSError(f'Failed reading secrets_dir {p!r}: {e}') from e

    return out


def dotenv_values(files):
    """
    Retrieve the values (environment variables) from a dotenv file,
    or a list/tuple of dotenv files.
    """
    if files is True:
        files = ['.env']
    elif isinstance(files, (str, PathLike)):
        files = [files]

    env = {}

    for f in files:
        f = fspath(f)

        # iterate backwards (from current directory) to find the
        # dotenv file
        #
        # If user gave a path (absolute or contains a separator), don't search.
        if isabs(f) or (sep in f) or (altsep and altsep in f):
            dotenv_path = f
        else:
            dotenv_path = dotenv.find_dotenv(f, usecwd=True)

        if not dotenv_path:  # not found
            continue

        # take environment variables from `.env` file
        env.update(dotenv.dotenv_values(dotenv_path))

    return env
